"""
@file
@brief Implements a custom criterion to train a decision tree.
"""
cimport cython
import numpy
cimport numpy

numpy.import_array()

from libc.stdlib cimport calloc, free
from libc.string cimport memcpy
from libc.math cimport NAN

cimport scipy.linalg.cython_lapack as cython_lapack
from sklearn.tree._criterion cimport SIZE_t, DOUBLE_t
from ._piecewise_tree_regression_common cimport CommonRegressorCriterion


cdef class LinearRegressorCriterion(CommonRegressorCriterion):
    """
    Criterion which computes the mean square error
    assuming points falling into one node are approximated
    by a line (linear regression).
    The implementation follows the same
    design used in :class:`SimpleRegressorCriterion
    <mlinsights.mlmodel.piecewise_tree_regression_criterion.SimpleRegressorCriterion>`
    and is even slow as the criterion is more complex to compute.
    """
    cdef DOUBLE_t* sample_w
    cdef DOUBLE_t* sample_y
    cdef DOUBLE_t* sample_wy
    cdef DOUBLE_t* sample_f
    cdef DOUBLE_t* sample_pC
    cdef DOUBLE_t* sample_pS
    cdef DOUBLE_t* sample_work
    cdef SIZE_t* sample_i
    cdef DOUBLE_t* sample_f_buffer
    
    cdef DOUBLE_t sample_sum_wy
    cdef DOUBLE_t sample_sum_w
    cdef SIZE_t nbvar
    cdef SIZE_t nbrows
    cdef SIZE_t work

    def __dealloc__(self):
        """Destructor."""        
        free(self.sample_w)
        free(self.sample_y)
        free(self.sample_wy)
        free(self.sample_i)
        free(self.sample_f)
        free(self.sample_f_buffer)
        free(self.sample_pC)
        free(self.sample_pS)
        free(self.sample_work)
        self.sample_w = NULL
        self.sample_y = NULL
        self.sample_wy = NULL
        self.sample_i = NULL
        self.sample_f = NULL
        self.sample_f_buffer = NULL
        self.sample_pC = NULL
        self.sample_pS = NULL
        self.sample_work = NULL

    def __getstate__(self):
        return {}

    def __setstate__(self, d):
        pass

    def __cinit__(self, const DOUBLE_t[:, ::1] X):
        self.sample_X = X
        # Allocate memory for the accumulators
        self.sample_w = NULL
        self.sample_y = NULL
        self.sample_wy = NULL
        self.sample_i = NULL
        self.sample_f = NULL
        self.sample_f_buffer = NULL
        self.sample_pC = NULL
        self.sample_pS = NULL
        self.sample_work = NULL
        
        # Criterion interface
        self.sample_weight = NULL
        self.samples = NULL
        self.sum_total = NULL
        self.sum_left = NULL
        self.sum_right = NULL            

        # allocation
        if self.sample_w == NULL:
            self.sample_w = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_wy == NULL:
            self.sample_wy = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_y == NULL:
            self.sample_y = <DOUBLE_t*> calloc(X.shape[0], sizeof(DOUBLE_t))
        if self.sample_i == NULL:
            self.sample_i = <SIZE_t*> calloc(X.shape[0], sizeof(SIZE_t))
        if self.sample_f == NULL:
            self.sample_f = <DOUBLE_t*> calloc(X.shape[0] * (X.shape[1] + 1), sizeof(DOUBLE_t))
            
        self.nbvar = X.shape[1] + 1
        self.nbrows = X.shape[0]
        self.work = <SIZE_t>(min(self.nbrows, self.nbvar) * <SIZE_t>3 + 
                             max(max(self.nbrows, self.nbvar),
                                 min(self.nbrows, self.nbvar) * <SIZE_t>2))
        if self.sample_f_buffer == NULL:
            self.sample_f_buffer = <DOUBLE_t*> calloc(X.shape[0] * self.nbvar, sizeof(DOUBLE_t))
        if self.sample_pC == NULL:
            self.sample_pC = <DOUBLE_t*> calloc(max(self.nbrows, self.nbvar), sizeof(DOUBLE_t))
        if self.sample_work == NULL:
            self.sample_work = <DOUBLE_t*> calloc(self.work, sizeof(DOUBLE_t))
        if self.sample_pS == NULL:
            self.sample_pS = <DOUBLE_t*> calloc(self.nbvar, sizeof(DOUBLE_t))

    @staticmethod
    def create(DOUBLE_t[:, ::1] X, DOUBLE_t[:, ::1] y, DOUBLE_t[::1] sample_weight=None):
        """
        Initializes the criterion.
        
        :param X: features
        :param y: target
        :param sample_weight: sample weight
        :return: an instance of :class:`LinearRegressorCriterion`
        """
        cdef SIZE_t i
        cdef DOUBLE_t* ws
        cdef double sum
        cdef SIZE_t* parr = <SIZE_t*> calloc(y.shape[0], sizeof(SIZE_t))
        for i in range(0, y.shape[0]):
            parr[i] = i
        if sample_weight is None:
            sum = <DOUBLE_t>y.shape[0]
            ws = NULL
        else:
            sum = sample_weight.sum()
            ws = &sample_weight[0]

        obj = LinearRegressorCriterion(X)
        obj.init(y, ws, sum, parr, 0, y.shape[0])            
        free(parr)
        return obj

    cdef int init(self, const DOUBLE_t[:, ::1] y,
                  DOUBLE_t* sample_weight,
                  double weighted_n_samples, SIZE_t* samples, 
                  SIZE_t start, SIZE_t end) nogil except -1:
        """
        This function is overwritten to check *y* and *X* size are the same.
        This API changed in 0.21.
        """
        if y.shape[0] != self.sample_X.shape[0]:
            raise ValueError("X.shape={} -- y.shape={}".format(self.sample_X.shape, y.shape))
        if y.shape[1] != 1:
            raise ValueError("This class only works for a single vector.")
        return self.init_with_X(self.sample_X, y, sample_weight, weighted_n_samples,
                                samples, start, end)

    cdef int init_with_X(self, const DOUBLE_t[:, ::1] X, 
                         const DOUBLE_t[:, ::1] y,
                         DOUBLE_t* sample_weight,
                         double weighted_n_samples, SIZE_t* samples, 
                         SIZE_t start, SIZE_t end) nogil except -1:
        """
        Initializes the criterion.

        :param X: array-like, features, dtype=DOUBLE_t
        :param y: array-like, dtype=DOUBLE_t
            y is a buffer that can store values for n_outputs target variables
        :param sample_weight: array-like, dtype=DOUBLE_t
            The weight of each sample
        :param weighted_n_samples: DOUBLE_t
            The total weight of the samples being considered
        :param samples: array-like, dtype=DOUBLE_t
            Indices of the samples in X and y, where samples[start:end]
            correspond to the samples in this node
        :param start: SIZE_t
            The first sample to be used on this node
        :param end: SIZE_t
            The last sample used on this node
        """
        cdef SIZE_t ki, ks, idx, c

        self.start = start
        self.pos = start
        self.end = end
        self.weighted_n_samples = weighted_n_samples
        self.y = y    

        self.sample_sum_wy = 0.
        self.sample_sum_w = 0.

        # Filling accumulators.
        idx = start * self.nbvar
        for ki in range(<int>start, <int>end):
            ks = samples[ki]
            self.sample_i[ki] = ks
            self.sample_w[ki] = sample_weight[ks] if sample_weight else 1.
            self.sample_wy[ki] = self.sample_w[ki] * y[ks, 0]
            self.sample_y[ki] = y[ks, 0]
            self.sample_sum_wy += y[ks, 0] * self.sample_w[ki]
            self.sample_sum_w += self.sample_w[ki]
            for c in range(0, <int>self.nbvar-1):
                self.sample_f[idx] = X[ks, c]
                idx += 1
            self.sample_f[idx] = 1.
            idx += 1

        self.weighted_n_node_samples = self.sample_sum_w
        self.reset()
        if self.weighted_n_node_samples == 0:
            raise ValueError(
                "self.weighted_n_node_samples is null, first weight is %r." % self.sample_w[0])
        return 0

    cdef void _mean(self, SIZE_t start, SIZE_t end, DOUBLE_t *mean, DOUBLE_t *weight) nogil:
        """
        Computes mean between *start* and *end*.
        """
        if start == end:
            mean[0] = 0.
            return
        cdef DOUBLE_t m = 0.
        cdef DOUBLE_t w = 0.
        cdef SIZE_t k
        for k in range(<int>start, <int>end):
            m += self.sample_wy[k]
            w += self.sample_w[k]
        weight[0] = w
        mean[0] = 0. if w == 0. else m / w
            
    cdef void _reglin(self, SIZE_t start, SIZE_t end, int low_rank) nogil:
        """
        Solves the linear regression between *start* and *end*
        assuming corresponding points are approximated by a line.
        *mean* is unused but could be if the inverse does not exist.
        The solution is the vector ``self.sample_pC[:self.nbvar]``.
        """
        cdef SIZE_t i, j, idx, pos
        cdef DOUBLE_t w
        cdef DOUBLE_t* sample_f_buffer = self.sample_f_buffer
        pos = 0
        for j in range(self.nbvar):
            idx = start * self.nbvar + j
            for i in range(<int>start, <int>end):
                w = self.sample_w[i]
                sample_f_buffer[pos] = self.sample_f[idx] * w
                idx += self.nbvar
                pos += 1
        
        cdef DOUBLE_t* pC = self.sample_pC
        for i in range(<int>start, <int>end):
            pC[i-start] = self.sample_wy[i]

        cdef int col = <int>self.nbvar
        cdef int row = <int>(end - start)
        cdef int info
        cdef int nrhs = 1
        cdef int lda = row
        cdef int ldb = row
        cdef DOUBLE_t rcond = -1
        cdef int rank        
        cdef int work = <int>self.work
        
        if row < col:
            if low_rank:
                ldb = col
            else:
                raise RuntimeError("The function cannot return any return when row < col.")
        cython_lapack.dgelss(&row, &col, &nrhs,                 # 1-3
                             sample_f_buffer, &lda, pC, &ldb,   # 4-7
                             self.sample_pS, &rcond, &rank,     # 8-10
                             self.sample_work, &work, &info)    # 11-13
                             
    cdef double _mse(self, SIZE_t start, SIZE_t end, DOUBLE_t mean, DOUBLE_t weight) nogil:
        """
        Computes mean square error between *start* and *end*
        assuming corresponding points are approximated by a line.
        *mean* is unused but could be if the inverse does not exist.
        """
        if end - start <= self.nbvar:
            # More coefficients than the number of observations.
            return 0.
        
        self._reglin(start, end, 0)
        
        cdef double* pC = self.sample_pC
        cdef SIZE_t j, idx
        
        # replaces what follows by gemm
        cdef DOUBLE_t squ = 0.
        cdef DOUBLE_t d
        cdef SIZE_t k
        idx = start * self.nbvar
        for k in range(<int>start, <int>end):
            d = 0.
            for j in range(self.nbvar):
                d += self.sample_f[idx] * pC[j]
                idx += 1
            d -= self.sample_y[k]
            squ += d * d * self.sample_w[k]
        return 0. if weight == 0. else squ / weight
        
    cdef void _node_beta(self, double* dest) nogil:
        """
        Stores the results of the linear regression
        in an allocated numpy array.
        
        :param dest: allocated double pointer, size must be *>= self.nbvar*
        """
        self._reglin(self.start, self.end, 1)
        memcpy(dest, self.sample_pC, self.nbvar * sizeof(double))

    def node_beta(self, double[::1] dest):
        """
        Stores the results of the linear regression
        in an allocated numpy array.
        
        :param dest: allocated array
        """
        if dest.shape[0] < self.nbvar:
            raise ValueError("dest must be at least (%d, )" % self.nbvar)
        self._node_beta(&dest[0])



# -*- coding: utf-8 -*-
"""
@file
@brief Implémente la classe @see cl ConstraintKMeans.
"""
import numpy
from scipy.spatial import Delaunay  # pylint: disable=E0611
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import euclidean_distances
from ._kmeans_constraint_ import constraint_kmeans, constraint_predictions


class ConstraintKMeans(KMeans):
    """
    Defines a constraint :epkg:`k-means`.
    Clusters are modified to have an equal size.
    The algorithm is initialized with a regular :epkg:`KMeans`
    and continues with a modified version of it.

    Computing the predictions offer a choice.
    The first one is to keep the predictions
    from the regular :epkg:`k-means`
    algorithm but with the balanced clusters.
    The second is to compute balanced predictions
    over the test set. That implies that the predictions
    for the same observations might change depending
    on the set it belongs to.

    The parameter *strategy* determines how
    obseervations should be assigned to a cluster.
    The value can be:

    * ``'distance'``: observations are ranked by distance to a cluster,
      the algorithm assigns first point to the closest center unless it reached
      the maximum size, it deals first with the further point and maps it to
      the closest center
    * ``'gain'``: follows the algorithm described at
       see `Same-size k-Means Variation
       <https://elki-project.github.io/tutorial/same-size_k_means>`_,
    * ``'weights'``: estimates weights attached to each cluster,
        it weights the distance to each cluster in order
        to balance the number of points mapped to every cluster,
        the strategy uses a learning rate.

    The first two strategies cannot reach a good compromise
    without using function @see fn _switch_clusters which
    tries every switch between clusters: two points
    change clusters. It keeps the number of points and checks
    that the inertia is reduced.
    """

    _strategy_value = {'distance', 'gain', 'weights'}

    def __init__(self, n_clusters=8, init='k-means++', n_init=10, max_iter=500,
                 tol=0.0001, precompute_distances='deprecated', verbose=0,
                 random_state=None, copy_x=True, n_jobs=1, algorithm='auto',
                 balanced_predictions=False, strategy='gain', kmeans0=True,
                 learning_rate=1., history=False):
        """
        @param      n_clusters              number of clusters
        @param      init                    used by :epkg:`k-means`
        @param      n_init                  used by :epkg:`k-means`
        @param      max_iter                used by :epkg:`k-means`
        @param      tol                     used by :epkg:`k-means`
        @param      precompute_distances    used by :epkg:`k-means`
        @param      verbose                 used by :epkg:`k-means`
        @param      random_state            used by :epkg:`k-means`
        @param      copy_x                  used by :epkg:`k-means`
        @param      n_jobs                  used by :epkg:`k-means`
        @param      algorithm               used by :epkg:`k-means`
        @param      balanced_predictions    produced balanced prediction
                                            or the regular ones
        @param      strategy                strategy or algorithm used to abide
                                            by the constraint
        @param      kmeans0                 if True, applies *k-means* algorithm first
        @param      history                 keeps centers accress iterations
        @param      learning_rate           learning rate, used by strategy `'weights'`
        """
        KMeans.__init__(self, n_clusters=n_clusters, init=init, n_init=n_init,
                        max_iter=max_iter, tol=tol, precompute_distances=precompute_distances,
                        verbose=verbose, random_state=random_state, copy_x=copy_x,
                        n_jobs=n_jobs, algorithm=algorithm)
        self.balanced_predictions = balanced_predictions
        self.strategy = strategy
        self.kmeans0 = kmeans0
        self.history = history
        self._n_threads = None
        self.learning_rate = learning_rate
        if strategy not in ConstraintKMeans._strategy_value:
            raise ValueError('strategy must be in {0}'.format(
                ConstraintKMeans._strategy_value))
        if precompute_distances == 'deprecated':
            km = KMeans()
            if km.precompute_distances != precompute_distances:
                self.precompute_distances = km.precompute_distances

    def fit(self, X, y=None, sample_weight=None, fLOG=None):
        """
        Compute k-means clustering.

        :param X: array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances to cluster. It must be noted that the data
            will be converted to C ordering, which will cause a memory
            copy if the given data is not C-contiguous.
        :param y: Ignored
        :param sample_weight: sample weight
        :param fLOG: logging function
        """
        max_iter = self.max_iter
        self.max_iter //= 2
        if self.kmeans0:
            KMeans.fit(self, X, y, sample_weight=sample_weight)
            state = None
        else:
            state = numpy.random.RandomState(  # pylint: disable=E1101
                self.random_state)
            labels = state.randint(
                0, self.n_clusters, X.shape[0], dtype=numpy.int32)
            centers = numpy.empty((self.n_clusters, X.shape[1]), dtype=X.dtype)
            choice = state.randint(0, self.n_clusters, self.n_clusters)
            for i, c in enumerate(choice):
                centers[i, :] = X[c, :]
            self.labels_ = labels
            self.cluster_centers_ = centers
            self.inertia_ = float(X.shape[0])
            self.n_iter_ = 0

        self.max_iter = max_iter
        return self.constraint_kmeans(
            X, sample_weight=sample_weight, state=state,
            learning_rate=self.learning_rate,
            history=self.history, fLOG=fLOG)

    def constraint_kmeans(self, X, sample_weight=None, state=None,
                          learning_rate=1., history=False, fLOG=None):
        """
        Completes the constraint k-means.

        @param      X               features
        @param      sample_weight   sample weight
        @param      state           state
        @param      history         keeps evolution of centers
        @param      fLOG            logging function
        """
        labels, centers, inertia, weights, iter_, all_centers = constraint_kmeans(
            X, self.labels_, sample_weight, self.cluster_centers_,
            inertia=self.inertia_, iter=self.n_iter_,
            max_iter=self.max_iter, verbose=self.verbose,
            strategy=self.strategy, state=state,
            learning_rate=learning_rate, history=history,
            fLOG=fLOG)
        self.labels_ = labels
        self.cluster_centers_ = centers
        self.cluster_centers_iter_ = (
            None if len(all_centers) == 0 else numpy.dstack(all_centers))
        self.inertia_ = inertia
        self.n_iter_ = iter_
        self.weights_ = weights
        return self

    def predict(self, X, sample_weight=None):
        """
        Computes the predictions.

        @param      X       features.
        @return             prediction
        """
        if self.weights_ is None:
            if self.balanced_predictions:
                labels, _, __ = constraint_predictions(
                    X, self.cluster_centers_, strategy=self.strategy + '_p')
                return labels
            return KMeans.predict(self, X, sample_weight=sample_weight)
        else:
            if self.balanced_predictions:
                raise RuntimeError(  # pragma: no cover
                    "balanced_predictions and weights_ cannot be used together.")
            return KMeans.predict(self, X, sample_weight=sample_weight)

    def transform(self, X):
        """
        Computes the predictions.

        @param      X       features.
        @return             prediction
        """
        if self.weights_ is None:
            if self.balanced_predictions:
                labels, distances, __ = constraint_predictions(
                    X, self.cluster_centers_, strategy=self.strategy)
                # We remove small distances than the chosen clusters
                # due to the constraint, we choose max*2 instead.
                mx = distances.max() * 2
                for i, l in enumerate(labels):
                    mi = distances[i, l]
                    mmi = distances[i, :].min()
                    if mi > mmi:
                        # numpy.nan would be best
                        distances[i, distances[i, :] < mi] = mx
                return distances
            return KMeans.transform(self, X)
        else:
            if self.balanced_predictions:
                raise RuntimeError(  # pragma: no cover
                    "balanced_predictions and weights_ cannot be used together.")
            res = KMeans.transform(self, X)
            res *= self.weights_.reshape((1, -1))
            return res

    def score(self, X, y=None, sample_weight=None):
        """
        Returns the distances to all clusters.

        @param      X               features
        @param      y               unused
        @param      sample_weight   sample weight
        @return                     distances
        """
        if self.weights_ is None:
            if self.balanced_predictions:
                _, __, dist_close = constraint_predictions(
                    X, self.cluster_centers_, strategy=self.strategy)
                return dist_close
            res = euclidean_distances(self.cluster_centers_, X, squared=True)
        else:
            if self.balanced_predictions:
                raise RuntimeError(  # pragma: no cover
                    "balanced_predictions and weights_ cannot be used together.")
            res = euclidean_distances(X, self.cluster_centers_, squared=True)
            res *= self.weights_.reshape((1, -1))
        return res.max(axis=1)

    def cluster_edges(self):
        """
        Computes edges between clusters based on a
        `Delaunay <https://docs.scipy.org/doc/scipy/reference/
        generated/scipy.spatial.Delaunay.html>`_
        graph.
        """
        tri = Delaunay(self.cluster_centers_)
        triangles = tri.simplices  # pylint: disable=E1101
        edges = set()
        for row in triangles:
            for j in range(1, row.shape[-1]):
                a, b = row[j - 1:j + 1]
                if a < b:
                    edges.add((a, b))
                else:
                    edges.add((b, a))
            a, b = row[0], row[-1]
            if a < b:
                edges.add((a, b))
            else:
                edges.add((b, a))
        return edges



"""
@file
@brief Direct calls to libraries :epkg:`BLAS` and :epkg:`LAPACK`.
"""
from libc.stdlib cimport calloc, free
from libc.string cimport memcpy
from libc.stdio cimport printf
from libc.math cimport NAN

import numpy
cimport numpy
cimport cython
numpy.import_array()
cimport scipy.linalg.cython_lapack as cython_lapack


@cython.boundscheck(False)
@cython.wraparound(False)
def dgelss(double[:, ::1] A, double [:, ::1] B, double prec=-1.):
    """
    Finds *X* in the problem :math:`AX=B` by minimizing
    :math:`\\norm{AX - B}^2`. Uses function
    `dgels <http://www.netlib.org/lapack/explore-html/d8/dde/dgels_8f.html>`_.
    
    :param A: matrix with 2 dimensions
    :param B: matrix with 2 dimensions
    :param prec: precision
    :return: integer (INFO)
    
    INFO is:
    
    * ``= 0``: successful exit
    * ``< 0``: if INFO = -i, the i-th argument had an illegal value
    * ``> 0``: if INFO =  i, the i-th diagonal element of the
      triangular factor of A is zero, so that A does not have
      full rank; the least squares solution could not be
      computed.
      
    .. note::
        ``::1`` indicates A, B, C must be contiguous arrays.
        Arrays *A*, *B* are modified by the function.
        *B* contains the solution.

    .. exref::
        :title: Use lapack function dgelss
        
        *C* minimizes the problem :math:`\\norm{AX - B}^2`.
        
        .. runpython::
            :showcode:
        
            import numpy
            from scipy.linalg.lapack import dgelss as scipy_dgelss
            from mlinsights.mlmodel.direct_blas_lapack import dgelss
        
            A = numpy.array([[10., 1.], [12., 1.], [13., 1]])
            B = numpy.array([[20., 22., 23.]]).T
            v, x, s, rank, work, info = scipy_dgelss(A, B)
            print(x[:2])
            
            A = A.T.copy()
            info = dgelss(A, B)
            assert info == 0
            print(B[:2])
    """
    if A.shape[1] != B.shape[0]:
        raise ValueError("A and B have mismatched dimensions: %d != %d." % (A.shape[1], B.shape[0]))
    cdef int res
    cdef int rank
    with nogil:
        res = _dgelss(A, B, &rank, &prec)
    return res
    
    
cdef void copy2array2(const double* pC, double[:, ::1] C) nogil:
    """
    Copies double from a buffer to an array.
    """
    cdef size_t size = C.shape[0] * C.shape[1]
    memcpy(&C[0,0], pC, size * sizeof(double))
    
                
cdef void copy2array1(const double* pC, double[::1] C) nogil:
    """
    Copies double from a buffer to an array.
    """
    cdef size_t size = C.shape[0]
    memcpy(&C[0], pC, size * sizeof(double))
    
                
cdef int _dgelss(double[:, ::1] A, double [:, ::1] B, int* rank, const double * rcond) nogil:
    """
    Same function as :func:`dgels` but does no check.
    """
    cdef int col = A.shape[0]
    cdef int row = A.shape[1]
    cdef int info
    cdef double *pC
    cdef double *pS
    cdef int work = min(row, col) * 3 + max(max(row, col), min(row, col) * 2)
    
    pC = <double*> calloc(work, sizeof(double))
    pS = <double*> calloc(col, sizeof(double))
    
    _dgelss_noalloc(A, B, rank, rcond, pS, pC, &work, &info)

    free(pC)
    free(pS)
    return info


cdef void _dgelss_noalloc(double[:, ::1] A, double [:, ::1] B, int* rank, const double* rcond,
                          double* pS, double *pC, int* work, int* info) nogil:
    """
    Same function as :func:`dgels` but does no check.
    """
    cdef int col = A.shape[0]
    cdef int row = A.shape[1]
    cdef int nrhs = B.shape[1]
    cdef int lda = row
    cdef int ldb = row
    
    cython_lapack.dgelss(&row, &col, &nrhs,             # 1-3
                         &A[0,0], &lda, &B[0,0], &ldb,  # 4-7
                         pS, rcond, rank,               # 8-10
                         pC, work, info)                # 11-13



"""
@file
@brief Builds a tree of logistic regressions.
"""
import numpy
import scipy.sparse as sparse
from sklearn.linear_model import LogisticRegression
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.linear_model._base import LinearClassifierMixin


def logistic(x):
    """
    Computes :math:`\\frac{1}{1 + e^{-x}}`.
    """
    return 1. / (1. + numpy.exp(-x))


def likelihood(x, y, theta=1., th=0.):
    """
    Computes :math:`\\sum_i y_i f(\\theta (x_i - x_0)) + (1 - y_i) (1 - f(\\theta (x_i - x_0)))`
    where :math:`f(x_i)` is :math:`\\frac{1}{1 + e^{-x}}`.
    """
    lr = logistic((x - th) * theta)
    return y * lr + (1. - y) * (1 - lr)


class _DecisionTreeLogisticRegressionNode:
    """
    Describes the tree structure hold by class
    @see cl DecisionTreeLogisticRegression.
    See also notebook :ref:`decisiontreelogregrst`.
    """

    def __init__(self, estimator, threshold=0.5, depth=1, index=0):
        """
        constructor

        @param      estimator       binary estimator
        """
        self.index = index
        self.estimator = estimator
        self.above = None
        self.below = None
        self.threshold = threshold
        self.depth = depth

    def predict(self, X):
        """
        Predicts
        """
        prob = self.predict_proba(X)
        return (prob[:, 1] >= 0.5).astype(numpy.int32)

    def predict_proba(self, X):
        """
        Returns the classification probabilities.

        @param          X   features
        @return             probabilties
        """
        prob = self.estimator.predict_proba(X)
        above = prob[:, 1] > self.threshold
        below = ~ above
        n_above = above.sum()
        n_below = below.sum()
        if self.above is not None and n_above > 0:
            prob_above = self.above.predict_proba(X[above])
            prob[above] = prob_above
        if self.below is not None and n_below > 0:
            prob_below = self.below.predict_proba(X[below])
            prob[below] = prob_below
        return prob

    def decision_path(self, X, mat, indices):
        """
        Returns the classification probabilities.

        @param          X       features
        @param          mat     decision path (allocated matrix)
        """
        mat[indices, self.index] = 1
        prob = self.estimator.predict_proba(X)
        above = prob[:, 1] > self.threshold
        below = ~ above
        n_above = above.sum()
        n_below = below.sum()
        indices_above = indices[above]
        indices_below = indices[below]
        if self.above is not None and n_above > 0:
            self.above.decision_path(X[above], mat, indices_above)
        if self.below is not None and n_below > 0:
            self.below.decision_path(X[below], mat, indices_below)

    def fit(self, X, y, sample_weight, dtlr, total_N):
        """
        Fits a logistic regression, then splits the sample into
        positive and negative examples, finally tries to fit
        logistic regressions on both subsamples. This method only
        works on a linear classifier.

        @param      X               features
        @param      y               binary labels
        @param      sample_weight   weights of every sample
        @param      dtlr            @see cl DecisionTreeLogisticRegression
        @param      total_N         total number of observation
        @return                     last index
        """
        self.estimator.fit(X, y, sample_weight=sample_weight)
        if dtlr.verbose >= 1:
            print("[DTLR ] %s trained acc %1.2f N=%d" % (  # pragma: no cover
                " " * self.depth, self.estimator.score(X, y), X.shape[0]))
        prob = self.fit_improve(dtlr, total_N, X, y,
                                sample_weight=sample_weight)

        if self.depth + 1 > dtlr.max_depth:
            return self.index
        if X.shape[0] < dtlr.min_samples_split:
            return self.index

        above = prob[:, 1] > self.threshold
        below = ~ above
        n_above = above.sum()
        n_below = below.sum()
        y_above = set(y[above])
        y_below = set(y[below])

        def _fit_side(index, y_above_below, above_below, n_above_below, side):
            if dtlr.verbose >= 1:
                print("[DTLR*] %s%s: n_class=%d N=%d - %d/%d" % (
                    " " * self.depth, side,
                    len(y_above_below), above_below.shape[0],
                    n_above_below, total_N))
            if (len(y_above_below) > 1 and
                    above_below.shape[0] > dtlr.min_samples_leaf * 2 and
                    (float(n_above_below) / total_N >=
                        dtlr.min_weight_fraction_leaf * 2) and
                    n_above_below < total_N):
                estimator = clone(dtlr.estimator)
                sw = sample_weight[above_below] if sample_weight is not None else None
                node = _DecisionTreeLogisticRegressionNode(
                    estimator, self.threshold, depth=self.depth + 1, index=index)
                last_index = node.fit(
                    X[above_below], y[above_below], sw, dtlr, total_N)
                return node, last_index
            return None, index

        self.above, last = _fit_side(
            self.index + 1, y_above, above, n_above, "above")
        self.below, last = _fit_side(
            last + 1, y_below, below, n_below, "below")
        return last

    @property
    def tree_depth_(self):
        """
        Returns the maximum depth of the tree.
        """
        dt = self.depth
        if self.above is not None:
            dt = max(dt, self.above.tree_depth_)
        if self.below is not None:
            dt = max(dt, self.below.tree_depth_)
        return dt

    def fit_improve(self, dtlr, total_N, X, y, sample_weight):
        """
        The method only works on a linear classifier, it changes
        the intercept in order to be within the constraints
        imposed by the *min_samples_leaf* and *min_weight_fraction_leaf*.
        The algorithm has a significant cost as it sorts every observation
        and chooses the best intercept.

        @param      dtlr            @see cl DecisionTreeLogisticRegression
        @param      total_N         total number of observations
        @param      X               features
        @param      y               labels
        @param      sample_weight   sample weight
        @return                     probabilities
        """
        if self.estimator is None:
            raise RuntimeError(
                "Estimator was not trained.")  # pragma: no cover
        prob = self.estimator.predict_proba(X)
        if dtlr.fit_improve_algo in (None, 'none'):
            return prob

        if not isinstance(self.estimator, LinearClassifierMixin):
            # The classifier is not linear and cannot be improved.
            if dtlr.fit_improve_algo == 'intercept_sort_always':  # pragma: no cover
                raise RuntimeError(
                    "The model is not linear ({}), "
                    "intercept cannot be improved.".format(self.estimator.__class__.__name__))
            return prob

        above = prob[:, 1] > self.threshold
        below = ~ above
        n_above = above.sum()
        n_below = below.sum()
        n_min = min(n_above, n_below)
        p1p2 = float(n_above * n_below) / X.shape[0] ** 2
        if dtlr.verbose >= 2:
            print("[DTLRI] %s imp %d <> %d, p1p2=%1.3f <> %1.3f" % (  # pragma: no cover
                " " * self.depth, n_min, dtlr.min_samples_leaf,
                p1p2, dtlr.p1p2))
        if (n_min >= dtlr.min_samples_leaf and
                float(n_min) / total_N >= dtlr.min_weight_fraction_leaf and
                p1p2 > dtlr.p1p2 and
                dtlr.fit_improve_algo != 'intercept_sort_always'):
            return prob

        coef = self.estimator.coef_
        decision_function = (X @ coef.T).ravel()
        order = numpy.argsort(decision_function, axis=0)
        begin = dtlr.min_samples_leaf

        sorted_df = decision_function[order]
        sorted_y = decision_function[order]
        N = sorted_y.shape[0]

        best = None
        besti = None
        beta_best = None
        for i in range(begin, N - begin):
            beta = - sorted_df[i]
            like = numpy.sum(likelihood(decision_function + beta, y)) / N
            w = float(i * (N - i)) / N**2
            like += w * dtlr.gamma
            if besti is None or like > best:
                best = like
                besti = i
                beta_best = beta

        if beta_best is not None:
            if dtlr.verbose >= 1:
                print("[DTLRI] %s change intercept %f --> %f in [%f, %f]" % (  # pragma: no cover
                    " " * self.depth, self.estimator.intercept_, beta_best,
                    - sorted_df[-1], - sorted_df[0]))
            self.estimator.intercept_ = beta_best
            prob = self.estimator.predict_proba(X)
        return prob

    def enumerate_leaves_index(self):
        """
        Returns the leaves index.
        """
        if self.above is None or self.below is None:
            yield self.index
        if self.above is not None:
            for index in self.above.enumerate_leaves_index():
                yield index
        if self.below is not None:
            for index in self.below.enumerate_leaves_index():
                yield index


class DecisionTreeLogisticRegression(BaseEstimator, ClassifierMixin):
    """
    Fits a logistic regression, then fits two other
    logistic regression for every observation on both sides
    of the border. It goes one until a tree is built.
    It only handles a binary classification.
    The built tree cannot be deeper than the maximum recursion.

    :param estimator: binary classification estimator,
        if empty, use a logistic regression, the theoritical
        model defined with a logistic regression but it could
        any binary classifier
    :param max_depth: int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples. It must be below the maximum
        allowed recursion by python.
    :param min_samples_split: int or float, default=2
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.
    :param min_samples_leaf: int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.
    :param min_weight_fraction_leaf: float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.
    :param fit_improve_algo: string, one of the following value:
        - `'auto'`: chooses the best option below, '`none'` for
          every non linear model, `'intercept_sort'` for linear models
        - '`none'`: does not nothing once the binary classifier is fit
        - `'intercept_sort'`: if one side of the classifier is too small,
          the method changes the best intercept possible verifying
          the constraints
        - `'intercept_sort_always'`: always chooses the best intercept
          possible
    :param p1p2: threshold in [0, 1]
        for every split, we can define probabilities :math:`p_1 p_2`
        which define the ratio of samples in both splits,
        if :math:`p_1 p_2` is below the threshold,
        method *fit_improve* is called
    :param gamma: weight before the coefficient :math:`p (1-p)`.
        When the model tries to improve the linear classifier,
        it looks a better intercept which maximizes the
        likelihood and verifies the constraints.
        In order to force the classifier to choose a value
        which splits the dataset into 2 almost equal folds,
        the function maximimes :math:`likelihood + \\gamma p (1 - p)`
        where *p* is the proportion of samples falling in the first
        fold.
    :param verbose: prints out information about the training

    Fitted attributes:

    *  `classes_`: ndarray of shape (n_classes,) or list of ndarray
        The classes labels (single output problem),
        or a list of arrays of class labels (multi-output problem).
    * `tree_`: Tree
        The underlying Tree object.
    """

    _fit_improve_algo_values = (
        None, 'none', 'auto', 'intercept_sort', 'intercept_sort_always')

    def __init__(self, estimator=None,
                 max_depth=20, min_samples_split=2,
                 min_samples_leaf=2, min_weight_fraction_leaf=0.0,
                 fit_improve_algo='auto', p1p2=0.09,
                 gamma=1., verbose=0):
        "constructor"
        ClassifierMixin.__init__(self)
        BaseEstimator.__init__(self)
        # logistic regression
        if estimator is None:
            self.estimator = LogisticRegression()
        else:
            self.estimator = estimator
        if max_depth is None:
            raise ValueError("'max_depth' cannot be None.")  # pragma: no cover
        if max_depth > 1024:
            raise ValueError(
                "'max_depth' must be <= 1024.")  # pragma: no cover
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.fit_improve_algo = fit_improve_algo
        self.p1p2 = p1p2
        self.gamma = gamma
        self.verbose = verbose

        if self.fit_improve_algo not in DecisionTreeLogisticRegression._fit_improve_algo_values:
            raise ValueError(
                "fit_improve_algo='{}' not in {}".format(
                    self.fit_improve_algo, DecisionTreeLogisticRegression._fit_improve_algo_values))

    def fit(self, X, y, sample_weight=None):
        """
        Builds the tree model.

        :param X: numpy array or sparse matrix of shape [n_samples,n_features]
            Training data
        :param y: numpy array of shape [n_samples, n_targets]
            Target values. Will be cast to X's dtype if necessary
        :param sample_weight: numpy array of shape [n_samples]
            Individual weights for each sample
        :return: self : returns an instance of self.

        Fitted attributes:

        * `classes_`: classes
        * `tree_`: tree structure, see @see cl _DecisionTreeLogisticRegressionNode
        * `n_nodes_`: number of nodes
        """
        if not isinstance(X, numpy.ndarray):
            if hasattr(X, 'values'):
                X = X.values
        if not isinstance(X, numpy.ndarray):
            raise TypeError("'X' must be an array.")
        if (sample_weight is not None and
                not isinstance(sample_weight, numpy.ndarray)):
            raise TypeError(
                "'sample_weight' must be an array.")  # pragma: no cover
        self.classes_ = numpy.array(sorted(set(y)))
        if len(self.classes_) != 2:
            raise RuntimeError(
                "The model only supports binary classification but labels are "
                "{}.".format(self.classes_))
        cls = (y == self.classes_[1]).astype(numpy.int32)
        estimator = clone(self.estimator)
        self.tree_ = _DecisionTreeLogisticRegressionNode(estimator, 0.5)
        self.n_nodes_ = self.tree_.fit(
            X, cls, sample_weight, self, X.shape[0]) + 1
        return self

    def predict(self, X):
        """
        Runs the predictions.
        """
        labels = self.tree_.predict(X)
        return numpy.take(self.classes_, labels)

    def predict_proba(self, X):
        """
        Converts predictions into probabilities.
        """
        return self.tree_.predict_proba(X)

    def decision_function(self, X):
        """
        Calls *decision_function*.
        """
        raise NotImplementedError(  # pragma: no cover
            "Decision function is not available for this model.")

    @property
    def tree_depth_(self):
        """
        Returns the maximum depth of the tree.
        """
        return self.tree_.tree_depth_

    def decision_path(self, X, check_input=True):
        """
        Returns the decision path.

        @param      X               inputs
        @param      check_input     unused
        @return                     sparse matrix
        """
        mat = sparse.lil_matrix((X.shape[0], self.n_nodes_), dtype=numpy.int32)
        self.tree_.decision_path(X, mat, numpy.arange(X.shape[0]))
        return sparse.csr_matrix(mat)

    def get_leaves_index(self):
        """
        Returns the index of every leave.
        """
        indices = self.tree_.enumerate_leaves_index()
        return numpy.array(sorted(indices))



# -*- coding: utf-8 -*-
"""
@file
@brief Shortcuts for *mltricks*.
"""
from .sklearn_base_classifier import SkBaseClassifier
from .sklearn_base_learner import SkBaseLearner
from .sklearn_base_regressor import SkBaseRegressor
from .sklearn_base_transform_learner import SkBaseTransformLearner
from .sklearn_base_transform_stacking import SkBaseTransformStacking
from .sklearn_parameters import SkLearnParameters, SkException



"""
@file
@brief Base class for timeseries.
"""
from sklearn.base import BaseEstimator, RegressorMixin, clone
from ..mlmodel.sklearn_transform_inv import BaseReciprocalTransformer
from .metrics import ts_mape
from .utils import check_ts_X_y, build_ts_X_y


class BaseReciprocalTimeSeriesTransformer(BaseReciprocalTransformer):
    """
    Base for all timeseries preprocessing
    automatically applied within a predictor.
    """

    def __init__(self, context_length=0):
        """
        @param      context_length  number of previous observations to
                                    build or rebuild the observations
        """
        BaseReciprocalTransformer.__init__(self)
        self.context_length = context_length

    def fit(self, X, y, sample_weight=None):
        """
        Stores the first values.
        """
        raise NotImplementedError("Should be overwritten.")  # pragma: no cover

    def transform(self, X, y, sample_weight=None, context=None):
        """
        Transforms both *X* and *y*.
        Returns *X* and *y*, returns
        *sample_weight* as well if not None.
        The context is used when the *y* series stored
        in the predictor is not related to the *y* series
        given to the *transform* method.
        """
        raise NotImplementedError("Should be overwritten.")  # pragma: no cover

    def get_fct_inv(self):
        """
        Returns the reverse tranform.
        """
        raise NotImplementedError("Should be overwritten.")  # pragma: no cover


class BaseTimeSeries(BaseEstimator):
    """
    Base class to build a predictor on timeseries.
    The class computes one or several predictions at each time,
    between *delay1* and *delay2*. It computes:
    :math:`\\hat{Y_{t+d} = f(Y_{t-1}, ..., Y_{t-p})}`
    with *d* in *[delay1, delay2[* and
    :math:`1 \\leqslant p \\leqslant past`.
    """

    def __init__(self, past=1, delay1=1, delay2=2,
                 use_all_past=False, preprocessing=None):
        """
        @param      past            values to use to predict
        @param      delay1          the model computes the first prediction for
                                    *time=t + delay1*
        @param      delay2          the model computes the last prediction for
                                    *time=t + delay2* excluded
        @param      use_all_past    use all past features, not only the timeseries
        @param      preprocessing   preprocessing to apply before predicting,
                                    only the timeseries itselves, it can be
                                    a difference, it must be of type
                                    @see cl BaseReciprocalTimeSeriesTransformer
        """
        self.past = past
        self.delay1 = delay1
        self.delay2 = delay2
        self.use_all_past = use_all_past
        self.preprocessing = preprocessing
        if self.delay1 < 1:
            raise ValueError("delay1 must be >= 1")  # pragma: no cover
        if self.delay2 <= self.delay1:
            raise ValueError("delay2 must be >= 1")  # pragma: no cover
        if self.past < 0:
            raise ValueError("past must be > 0")  # pragma: no cover
        if (preprocessing is not None and
                not isinstance(preprocessing, BaseReciprocalTimeSeriesTransformer)):
            raise TypeError(  # pragma: no cover
                "preprocessing must be of type 'BaseReciprocalTimeSeriesTransformer' "
                "not {}".format(type(preprocessing)))

    def _fit_preprocessing(self, X, y, sample_weight=None):
        """
        Applies the preprocessing.
        *X*, *y*, *sample_weight*.

        :param X: output of
            X may be empty (None)
        :param y: timeseries (one single vector), array [n_obs]
        :param sample_weight: weights None or array [n_obs]
        :return: *X*, *y*, *sample_weight*
        """
        check_ts_X_y(self, X, y)

        if self.preprocessing is not None:
            self.preprocessing_ = clone(self.preprocessing)
            self.preprocessing_.fit(X, y, sample_weight)
            xyw = self.preprocessing_.transform(X, y, sample_weight)
            X, y = xyw[:2]
            sample_weight = xyw[-1] if sample_weight is not None else None
        return X, y, sample_weight

    def _base_fit_predict(self, X, y, sample_weight=None):
        """
        Trains the preprocessing and returns the modified
        *X*, *y*, *sample_weight*.

        :param X: output of
            X may be empty (None)
        :param y: timeseries (one single vector), array [n_obs]
        :param sample_weight: weights None or array [n_obs]
        :return: *X*, *y*, *sample_weight*

        The *y* series is moved by *self.delay1* in the past.
        """
        if y is None:
            raise RuntimeError("y cannot be None")
        X, y, sample_weight = build_ts_X_y(
            self, X, y, sample_weight, same_rows=True)
        X, y, sample_weight = self._fit_preprocessing(X, y, sample_weight)
        return X, y, sample_weight

    def has_preprocessing(self):
        """
        Tells if there is one preprocessing.
        """
        return hasattr(self, 'preprocessing_') and self.preprocessing_ is not None

    def _applies_preprocessing(self, X, y, sample_weight):
        """
        Applies the preprocessing to the series.
        """
        if self.has_preprocessing():
            xyw = self.preprocessing_.transform(X, y, sample_weight)
            X, y = xyw[:2]
            sample_weight = xyw[-1] if sample_weight is not None else None
        return X, y, sample_weight

    def _applies_preprocessing_inv(self, X, y, sample_weight):
        """
        Applies the preprocessing to the series.
        """
        if self.has_preprocessing():
            inv = self.preprocessing_.get_fct_inv()
            X, y, sample_weight = inv.transform(X, y, sample_weight)

        return X, y, sample_weight


class TimeSeriesRegressorMixin(RegressorMixin):
    """
    Addition to :epkg:`sklearn:base:RegressorMixin`.
    """

    def score(self, X, y, sample_weight=None):
        """
        Scores the prediction using
        @see fn ts_mape

        :param X: features
        :param y: expected values
        :param sample_weight: sample weight
        :return: see @see fn ts_mape
        """
        pred = self.predict(X, y)
        return ts_mape(y, pred, sample_weight=sample_weight)



"""
@file
@brief Caches to cache training.
"""
import numpy

_caches = {}


class MLCache:
    """
    Implements a cache to reduce the number of trainings
    a grid search has to do.
    """

    def __init__(self, name):
        """
        @param      name        name of the cache
        """
        self.name = name
        self.cached = {}
        self.count_ = {}

    def cache(self, params, value):
        """
        Caches one object.

        @param      params  dictionary of parameters
        @param      value   value to cache
        """
        key = MLCache.as_key(params)
        if key in self.cached:
            raise KeyError(  # pragma: no cover
                "Key {0} already exists".format(params))
        self.cached[key] = value
        self.count_[key] = 0

    def get(self, params, default=None):
        """
        Retrieves an element from the cache.

        @param      params  dictionary of parameters
        @param      default if not found
        @return             value or None if it does not exists
        """
        key = MLCache.as_key(params)
        res = self.cached.get(key, default)
        if res != default:
            self.count_[key] += 1
        return res

    def count(self, params):
        """
        Retrieves the number of times
        an elements was retrieved from the cache.

        @param      params  dictionary of parameters
        @return             int
        """
        key = MLCache.as_key(params)
        return self.count_.get(key, 0)

    @staticmethod
    def as_key(params):
        """
        Converts a list of parameters into a key.

        @param      params      dictionary
        @return                 key as a string
        """
        if isinstance(params, str):
            return params
        els = []
        for k, v in sorted(params.items()):
            if isinstance(v, (int, float, str)):
                sv = str(v)
            elif isinstance(v, tuple):
                if not all(map(lambda e: isinstance(e, (int, float, str)), v)):
                    raise TypeError(  # pragma: no cover
                        "Unable to create a key with value '{0}':{1}".format(k, v))
                return str(v)
            elif isinstance(v, numpy.ndarray):
                # id(v) may have been better but
                # it does not play well with joblib.
                sv = hash(v.tostring())
            elif v is None:
                sv = ""
            else:
                raise TypeError(  # pragma: no cover
                    "Unable to create a key with value '{0}':{1}".format(k, v))
            els.append((k, sv))
        return str(els)

    def __len__(self):
        """
        Returns the number of cached items.
        """
        return len(self.cached)

    def items(self):
        """
        Enumerates all cached items.
        """
        for item in self.cached.items():
            yield item

    def keys(self):
        """
        Enumerates all cached keys.
        """
        for k in self.cached.keys():  # pylint: disable=C0201
            yield k

    @staticmethod
    def create_cache(name):
        """
        Creates a new cache.

        @param      name        name
        @return                 created cache
        """
        global _caches  # pylint: disable=W0603
        if name in _caches:
            raise RuntimeError(  # pragma: no cover
                "cache '{0}' already exists.".format(name))

        cache = MLCache(name)
        _caches[name] = cache
        return cache

    @staticmethod
    def remove_cache(name):
        """
        Removes a cache with a given name.

        @param      name        name
        """
        global _caches  # pylint: disable=W0603
        del _caches[name]

    @staticmethod
    def get_cache(name):
        """
        Gets a cache with a given name.

        @param      name        name
        @return                 created cache
        """
        global _caches  # pylint: disable=W0603
        return _caches[name]

    @staticmethod
    def has_cache(name):
        """
        Tells if cache *name* is present.

        @param      name        name
        @return                 boolean
        """
        global _caches  # pylint: disable=W0603
        return name in _caches



"""
@file
@brief Featurizers for machine learned models.
"""
import numpy
import pandas
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier


class FeaturizerTypeError(TypeError):
    """
    Unable to process a type.
    """
    pass


def model_featurizer(model, **params):
    """
    Converts a machine learned model into a function which converts
    a vector into features produced by the model.
    It can be the output itself or intermediate results.
    The model can come from :epkg:`scikit-learn`,
    :epkg:`keras` or :epkg:`torch`.

    @param      model       model
    @param      params      additional parameters
    @return                 function
    """
    tried = []
    if isinstance(model, LogisticRegression):
        return model_featurizer_lr(model, **params)
    tried.append(LogisticRegression)
    if isinstance(model, RandomForestClassifier):
        return model_featurizer_rfc(model, **params)
    tried.append(RandomForestClassifier)
    if hasattr(model, "layers"):
        # It should be a keras model.
        return model_featurizer_keras(model, **params)
    tried.append("Keras")
    if hasattr(model, "forward"):
        # It should be a torch model.
        return model_featurizer_torch(model, **params)
    tried.append("torch")
    raise FeaturizerTypeError(  # pragma no cover
        "Unable to process type '{0}', allowed:\n{1}".format(
            type(model), "\n".join(sorted(str(_) for _ in tried))))


def is_vector(X):
    """
    Tells if *X* is a vector.

    @param      X       vector
    @return             boolean
    """
    if isinstance(X, list):
        if len(X) == 0 or isinstance(X[0], (list, tuple)):
            return False
        return True
    if isinstance(X, numpy.ndarray):
        if len(X.shape) > 1 and X.shape[0] != 1:
            return False
        return True
    if isinstance(X, pandas.DataFrame):
        if len(X.shape) > 1 and X.shape[0] != 1:
            return False
        return True
    raise TypeError(  # pragma no cover
        "Unable to guess if X is a vector, type(X)={0}".format(type(X)))


def wrap_predict_sklearn(X, fct, many):
    """
    Checks types and dimension.
    Calls *fct* and returns the approriate type.
    A vector if *X* is a vector, the raw output
    otherwise.

    @param      X       vector or list
    @param      fct     function
    @param      many    many observations or just one
    """
    isv = is_vector(X)
    if many == isv:
        raise ValueError(  # pragma: no cover
            "Inconsistency X is a single vector, many is True")
    if isv:
        X = [X]
    y = fct(X)
    if isv:
        y = y.ravel()
    return y


def model_featurizer_lr(model):
    """
    Builds a featurizer from a :epkg:`scikit-learn:linear_model:LogisticRegression`.
    It returns a function which returns ``model.decision_function(X)``.

    @param      model       model to use to featurize a vector
    @return                 function
    """

    def feat(X, model, many):
        "wraps sklearn"
        return wrap_predict_sklearn(X, model.decision_function, many)

    return lambda X, many, model=model: feat(X, model, many)


def model_featurizer_rfc(model, output=True):
    """
    Builds a featurizer from a :epkg:`scikit-learn:ensemble:RandomForestClassifier`.
    It returns a function which returns the output of every tree
    (method *apply*).

    @param      model       model to use to featurize a vector
    @param      output      use output (``model.predict_proba(X)``)
                            or trees output (``model.apply(X)``)
    @return                 function
    """
    if output:
        def feat1(X, model, many):
            "wraps sklearn"
            return wrap_predict_sklearn(X, model.predict_proba, many)

        return lambda X, many, model=model: feat1(X, model, many)

    def feat2(X, model, many):
        "wraps sklearn"
        return wrap_predict_sklearn(X, model.apply, many)

    return lambda X, many, model=model: feat2(X, model, many)


def wrap_predict_keras(X, fct, many, shapes):
    """
    Checks types and dimension.
    Calls *fct* and returns the approriate type.
    A vector if *X* is a vector, the raw output
    otherwise.

    @param      X       vector or list
    @param      fct     function
    @param      many    many observations or just one
    @param      shapes  expected input shapes for the neural network
    """
    if many:
        y = [fct(X[i]).ravel() for i in range(X.shape[0])]
        return numpy.stack(y)
    if len(X.shape) == len(shapes):
        return fct(X).ravel()
    x = X[numpy.newaxis, :, :, :]
    return fct(x).ravel()


def model_featurizer_keras(model, layer=None):
    """
    Builds a featurizer from a :epkg:`keras` model
    It returns a function which returns the output of one
    particular layer.

    @param      model       model to use to featurize a vector
    @param      layer       number of layers to keep
    @return                 function

    See `About Keras models <https://keras.io/models/about-keras-models/>`_.
    """
    if layer is not None:
        output = model.layers[layer].output
        model = model.__class__(model.input, output)

    def feat(X, model, many, shapes):
        "wraps keras"
        return wrap_predict_keras(X, model.predict, many, shapes)

    return lambda X, many, model=model, shapes=model._feed_input_shapes[0]: feat(X, model, many, shapes)


def wrap_predict_torch(X, fct, many, shapes):
    """
    Checks types and dimension.
    Calls *fct* and returns the approriate type.
    A vector if *X* is a vector, the raw output
    otherwise.

    @param      X       vector or list
    @param      fct     function
    @param      many    many observations or just one
    @param      shapes  expected input shapes for the neural network
    """
    if many:
        y = [fct(X[i]).ravel() for i in range(X.shape[0])]
        return numpy.stack(y)
    if shapes is None or len(X.shape) == len(shapes):
        t = fct(X)
        nt = t.detach().numpy().ravel()
    else:
        x = X[numpy.newaxis, :, :, :]
        t = fct(x)
        nt = t.detach().numpy().ravel()
    return nt


def model_featurizer_torch(model, layer=None):
    """
    Builds a featurizer from a :epkg:`torch` model
    It returns a function which returns the output of one
    particular layer.

    @param      model       model to use to featurize a vector
    @param      layer       number of layers to keep
    @return                 function
    """
    if layer is not None:
        output = model.layers[layer].output
        model = model.__class__(model.input, output)

    def feat(X, model, many, shapes):
        "wraps torch"
        return wrap_predict_torch(X, model.forward, many, shapes)

    return lambda X, many, model=model, shapes=None: feat(X, model, many, shapes)



"""
@file
@brief Implements a base class which defines a pair of transforms
applied around a predictor to modify the target as well.
"""
from sklearn.base import TransformerMixin, BaseEstimator


class BaseReciprocalTransformer(BaseEstimator, TransformerMixin):
    """
    Base for transform which transforms the features
    and the targets at the same time. It must also
    return another transform which transforms the target
    back to what it was.
    """

    def __init__(self):
        BaseEstimator.__init__(self)
        TransformerMixin.__init__(self)

    def get_fct_inv(self):
        """
        Returns a trained transform which reverse the target
        after a predictor.
        """
        raise NotImplementedError(
            "This must be overwritten.")  # pragma: no cover

    def transform(self, X, y):
        """
        Transforms *X* and *y*.
        Returns transformed *X* and *y*.
        """
        raise NotImplementedError(
            "This must be overwritten.")  # pragma: no cover



"""
@file
@brief Caches training.
"""
from distutils.version import StrictVersion
from sklearn import __version__ as skl_version
from sklearn.base import clone
from sklearn.pipeline import Pipeline, _fit_transform_one
from sklearn.utils import _print_elapsed_time
from .cache_model import MLCache


def isskl023():
    "Tells if :epkg:`scikit-learn` is more recent than 0.23."
    v1 = ".".join(skl_version.split('.')[:2])
    return StrictVersion(v1) >= StrictVersion('0.23')


class PipelineCache(Pipeline):
    """
    Same as :epkg:`sklearn:pipeline:Pipeline` but it can
    skip training if it detects a step was already trained
    the model was already trained accross
    even in a different pipeline.

    :param steps: list
        List of (name, transform) tuples (implementing fit/transform) that are
        chained, in the order in which they are chained, with the last object
        an estimator.
    :param cache_name: name of the cache, if None, a new name is created
    :param verbose: boolean, optional
        If True, the time elapsed while fitting each step will be printed as it
        is completed.

    Other attributes:

    :param named_steps: bunch object, a dictionary with attribute access
        Read-only attribute to access any step parameter by user given name.
        Keys are step names and values are steps parameters.
    """

    def __init__(self, steps, cache_name=None, verbose=False):
        self.cache_name = cache_name
        Pipeline.__init__(self, steps, memory=None, verbose=verbose)
        if cache_name is None:
            cache_name = "Pipeline%d" % id(self)
        self.cache_name = cache_name

    def _get_fit_params_steps(self, fit_params):
        fit_params_steps = {name: {} for name, step in self.steps
                            if step is not None}

        for pname, pval in fit_params.items():
            if '__' not in pname:
                if not isinstance(pval, dict):
                    raise ValueError(  # pragma: no cover
                        "For scikit-learn < 0.23, "
                        "Pipeline.fit does not accept the {} parameter. "
                        "You can pass parameters to specific steps of your "
                        "pipeline using the stepname__parameter format, e.g. "
                        "`Pipeline.fit(X, y, logisticregression__sample_weight"
                        "=sample_weight)`.".format(pname))
                else:
                    fit_params_steps[pname].update(pval)
            else:
                step, param = pname.split('__', 1)
                fit_params_steps[step][param] = pval
        return fit_params_steps

    def _fit(self, X, y=None, **fit_params):

        self.steps = list(self.steps)
        self._validate_steps()
        fit_params_steps = self._get_fit_params_steps(fit_params)
        if not MLCache.has_cache(self.cache_name):
            self.cache_ = MLCache.create_cache(self.cache_name)
        else:
            self.cache_ = MLCache.get_cache(self.cache_name)
        Xt = X
        for (step_idx, name, transformer) in self._iter(
                with_final=False, filter_passthrough=False):
            if (transformer is None or transformer == 'passthrough'):
                with _print_elapsed_time('Pipeline', self._log_message(step_idx)):
                    continue

            params = transformer.get_params()
            params['__class__'] = transformer.__class__.__name__
            params['X'] = Xt
            if ((hasattr(transformer, 'is_classifier') and transformer.is_classifier()) or
                    (hasattr(transformer, 'is_regressor') and transformer.is_regressor())):
                params['y'] = y
            cached = self.cache_.get(params)
            if cached is None:
                cloned_transformer = clone(transformer)
                Xt, fitted_transformer = _fit_transform_one(
                    cloned_transformer, Xt, y, None,
                    message_clsname='PipelineCache',
                    message=self._log_message(step_idx),
                    **fit_params_steps[name])
                self.cache_.cache(params, fitted_transformer)
            else:
                fitted_transformer = cached
                Xt = fitted_transformer.transform(Xt)

            self.steps[step_idx] = (name, fitted_transformer)
        if isskl023():
            return Xt
        if self._final_estimator == 'passthrough':
            return Xt, {}
        return Xt, fit_params_steps[self.steps[-1][0]]



# -*- coding: utf-8 -*-
"""
@file
@brief Implémente la classe @see cl ConstraintKMeans.
"""
import bisect
from collections import Counter
from pandas import DataFrame
import numpy
import scipy.sparse
from scipy.spatial.distance import cdist
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.utils.extmath import row_norms
from ._kmeans_022 import (
    _centers_dense, _centers_sparse, _labels_inertia_skl)


def linearize_matrix(mat, *adds):
    """
    Linearizes a matrix into a new one
    with 3 columns value, row, column.
    The output format is similar to
    :epkg:`csr_matrix` but null values are kept.

    @param      mat     matrix
    @param      adds    additional square matrices
    @return             new matrix

    *adds* defines additional matrices, it adds
    columns on the right side and fill them with
    the corresponding value taken into the additional
    matrices.
    """
    if scipy.sparse.issparse(mat):
        if isinstance(mat, scipy.sparse.csr_matrix):
            max_row = mat.shape[0]
            res = numpy.empty((len(mat.data), 3 + len(adds)), dtype=mat.dtype)
            row = 0
            for i, v in enumerate(mat.data):
                while row < max_row and i >= mat.indptr[row]:
                    row += 1
                res[i, 0] = v
                a, b = row - 1, mat.indices[i]
                res[i, 1] = a
                res[i, 2] = b
                for k, am in enumerate(adds):
                    res[i, k + 3] = am[a, b]
            return res
        raise NotImplementedError(  # pragma: no cover
            "This kind of sparse matrix is not handled: {0}".format(type(mat)))
    else:
        n = mat.shape[0]
        c = mat.shape[1]
        ic = numpy.arange(mat.shape[1])
        res = numpy.empty((n * c, 3 + len(adds)), dtype=mat.dtype)
        for i in range(0, n):
            a = i * c
            b = (i + 1) * c
            res[a:b, 1] = i
            res[a:b, 2] = ic
        res[:, 0] = mat.ravel()
        for k, am in enumerate(adds):
            res[:, 3 + k] = am.ravel()
        return res


def constraint_kmeans(X, labels, sample_weight, centers, inertia,
                      iter, max_iter,  # pylint: disable=W0622
                      strategy='gain', verbose=0, state=None,
                      learning_rate=1., history=False, fLOG=None):
    """
    Completes the constraint :epkg:`k-means`.

    @param      X               features
    @param      labels          initialized labels (unused)
    @param      sample_weight   sample weight
    @param      centers         initialized centers
    @param      inertia         initialized inertia (unused)
    @param      iter            number of iteration already done
    @param      max_iter        maximum of number of iteration
    @param      strategy        strategy used to sort observations before
                                mapping them to clusters
    @param      verbose         verbose
    @param      state           random state
    @param      learning_rate   used by strategy `'weights'`
    @param      history         return list of centers accross iterations
    @param      fLOG            logging function (needs to be specified otherwise
                                verbose has no effects)
    @return                     tuple (best_labels, best_centers, best_inertia,
                                iter, all_centers)
    """
    if labels.dtype != numpy.int32:
        raise TypeError(  # pragma: no cover
            "Labels must be an array of int not '{0}'".format(labels.dtype))

    if strategy == 'weights':
        return _constraint_kmeans_weights(
            X, labels, sample_weight, centers, inertia, iter,
            max_iter, verbose=verbose, state=state,
            learning_rate=learning_rate, history=history, fLOG=fLOG)
    else:
        if isinstance(X, DataFrame):
            X = X.values
        x_squared_norms = row_norms(X, squared=True)
        counters = numpy.empty((centers.shape[0],), dtype=numpy.int32)
        limit = X.shape[0] // centers.shape[0]
        leftover = X.shape[0] - limit * centers.shape[0]
        leftclose = numpy.empty((centers.shape[0],), dtype=numpy.int32)
        n_clusters = centers.shape[0]
        distances_close = numpy.empty((X.shape[0],), dtype=X.dtype)
        best_inertia = None
        best_iter = None
        all_centers = []

        # association
        _constraint_association(leftover, counters, labels, leftclose, distances_close,
                                centers, X, x_squared_norms, limit, strategy, state=state)

        if sample_weight is None:
            sw = numpy.ones((X.shape[0],))
        else:
            sw = sample_weight

        if scipy.sparse.issparse(X):
            _centers_fct = _centers_sparse
        else:
            _centers_fct = _centers_dense

        while iter < max_iter:
            # compute new clusters
            centers = _centers_fct(
                X, sw, labels, n_clusters, distances_close)

            if history:
                all_centers.append(centers)

            # association
            _constraint_association(
                leftover, counters, labels, leftclose, distances_close,
                centers, X, x_squared_norms, limit, strategy, state=state)

            # inertia
            _, inertia = _labels_inertia_skl(
                X=X, sample_weight=sw, x_squared_norms=x_squared_norms,
                centers=centers, distances=distances_close)

            iter += 1
            if verbose and fLOG:  # pragma: no cover
                fLOG("CKMeans %d/%d inertia=%f" % (iter, max_iter, inertia))

            # best option so far?
            if best_inertia is None or inertia < best_inertia:
                best_inertia = inertia
                best_centers = centers.copy()
                best_labels = labels.copy()
                best_iter = iter

            # early stop
            if (best_inertia is not None and inertia >= best_inertia and
                    iter > best_iter + 5):
                break

        return (best_labels, best_centers, best_inertia, None,
                iter, all_centers)


def constraint_predictions(X, centers, strategy, state=None):
    """
    Computes the predictions but tries
    to associates the same numbers of points
    in each cluster.

    @param      X           features
    @param      centers     centers of each clusters
    @param      strategy    strategy used to sort point before
                            mapping them to a cluster
    @param      state       random state
    @return                 labels, distances, distances_close
    """
    if isinstance(X, DataFrame):
        X = X.values
    x_squared_norms = row_norms(X, squared=True)
    counters = numpy.empty((centers.shape[0],), dtype=numpy.int32)
    limit = X.shape[0] // centers.shape[0]
    leftover = X.shape[0] - limit * centers.shape[0]
    leftclose = numpy.empty((centers.shape[0],), dtype=numpy.int32)
    distances_close = numpy.empty((X.shape[0],), dtype=X.dtype)
    labels = numpy.empty((X.shape[0],), dtype=numpy.int32)

    distances = _constraint_association(
        leftover, counters, labels, leftclose,
        distances_close, centers, X, x_squared_norms,
        limit, strategy, state=state)

    return labels, distances, distances_close


def _constraint_association(leftover, counters, labels, leftclose, distances_close,
                            centers, X, x_squared_norms, limit, strategy, state=None):
    """
    Completes the constraint :epkg:`k-means`.

    @param      X               features
    @param      labels          initialized labels (unused)
    @param      centers         initialized centers
    @param      x_squared_norms norm of *X*
    @param      limit           number of point to associate per cluster
    @param      leftover        number of points to associate at the end
    @param      counters        allocated array
    @param      leftclose       allocated array
    @param      labels          allocated array
    @param      distances_close allocated array
    @param      strategy        strategy used to sort point before
                                mapping them to a cluster
    @param      state           random state
    """
    if strategy in ('distance', 'distance_p'):
        return _constraint_association_distance(
            leftover, counters, labels, leftclose, distances_close,
            centers, X, x_squared_norms, limit, strategy, state=state)
    if strategy in ('gain', 'gain_p'):
        return _constraint_association_gain(
            leftover, counters, labels, leftclose, distances_close,
            centers, X, x_squared_norms, limit, strategy, state=state)
    raise ValueError("Unknwon strategy '{0}'.".format(
        strategy))  # pragma: no cover


def _compute_strategy_coefficient(distances, strategy, labels):
    """
    Creates a matrix
    """
    if strategy in ('gain', 'gain_p'):
        ar = numpy.arange(distances.shape[0])
        dist = distances[ar, labels]
        return distances - dist[:, numpy.newaxis]
    raise ValueError("Unknwon strategy '{0}'.".format(strategy))


def _randomize_index(index, weights):
    """
    Randomizes index depending on the value.
    Swap indexes.
    Modifies *index*.
    """
    maxi = weights.max()
    mini = weights.min()
    diff = max(maxi - mini, 1e-5)
    rand = numpy.random.rand(weights.shape[0])
    for i in range(1, index.shape[0]):
        ind1 = index[i - 1]
        ind2 = index[i]
        w1 = weights[ind1]
        w2 = weights[ind2]
        ratio = abs(w2 - w1) / diff * 0.5
        if rand[i] >= ratio + 0.5:
            index[i - 1], index[i] = ind2, ind1
            weights[i - 1], weights[i] = w2, w1


def _switch_clusters(labels, distances):
    """
    Tries to switch clusters.
    Modifies *labels* inplace.

    @param      labels      labels
    @param      distances   distances
    """
    perm = numpy.random.permutation(numpy.arange(0, labels.shape[0]))
    niter = 0
    modif = 1
    while modif > 0 and niter < 10:
        modif = 0
        niter += 1
        for i_ in range(labels.shape[0]):
            for j_ in range(i_ + 1, labels.shape[0]):
                i = perm[i_]
                j = perm[j_]
                c1 = labels[i]
                c2 = labels[j]
                if c1 == c2:
                    continue
                d11 = distances[i, c1]
                d12 = distances[i, c2]
                d21 = distances[j, c1]
                d22 = distances[j, c2]
                if d11**2 + d22**2 > d21**2 + d12**2:
                    labels[i], labels[j] = c2, c1
                    modif += 1


def _constraint_association_distance(leftover, counters, labels, leftclose, distances_close,
                                     centers, X, x_squared_norms, limit, strategy, state=None):
    """
    Completes the constraint *k-means*,
    the function sorts points by distance to the closest
    cluster and associates them into that order.
    It deals first with the further point and maps it to
    the closest center.

    @param      X               features
    @param      labels          initialized labels (unused)
    @param      centers         initialized centers
    @param      x_squared_norms norm of *X*
    @param      limit           number of point to associate per cluster
    @param      leftover        number of points to associate at the end
    @param      counters        allocated array
    @param      leftclose       allocated array
    @param      labels          allocated array
    @param      distances_close allocated array
    @param      strategy        strategy used to sort point before
                                mapping them to a cluster
    @param      state           random state (unused)
    """

    # initialisation
    counters[:] = 0
    leftclose[:] = -1
    labels[:] = -1

    # distances
    distances = euclidean_distances(
        centers, X, Y_norm_squared=x_squared_norms, squared=True)
    distances = distances.T
    distances0 = distances.copy()
    maxi = distances.ravel().max() * 2
    centers_index = numpy.argsort(distances, axis=1)

    while labels.min() == -1:
        mini = numpy.min(distances, axis=1)
        sorted_index = numpy.argsort(mini)
        _randomize_index(sorted_index, mini)

        nover = leftover
        for ind in sorted_index:
            if labels[ind] >= 0:
                continue
            for c in centers_index[ind, :]:
                if counters[c] < limit:
                    # The cluster still accepts new points.
                    counters[c] += 1
                    labels[ind] = c
                    distances[ind, c] = maxi
                    break
                if nover > 0 and leftclose[c] == -1:
                    # The cluster may accept one point if the number
                    # of clusters does not divide the number of points in X.
                    counters[c] += 1
                    labels[ind] = c
                    nover -= 1
                    leftclose[c] = 0
                    distances[ind, c] = maxi
                    break

    _switch_clusters(labels, distances0)
    distances_close[:] = distances[numpy.arange(X.shape[0]), labels]
    return distances0


def _constraint_association_gain(leftover, counters, labels, leftclose, distances_close,
                                 centers, X, x_squared_norms, limit, strategy, state=None):
    """
    Completes the constraint *k-means*.
    Follows the method described in `Same-size k-Means Variation
    <https://elki-project.github.io/tutorial/same-size_k_means>`_.

    @param      X               features
    @param      labels          initialized labels (unused)
    @param      centers         initialized centers
    @param      x_squared_norms norm of *X*
    @param      limit           number of points to associate per cluster
    @param      leftover        number of points to associate at the end
    @param      counters        allocated array
    @param      leftclose       allocated array
    @param      labels          allocated array
    @param      distances_close allocated array
    @param      strategy        strategy used to sort point before
                                mapping them to a cluster
    @param      state           random state

    See `Same-size k-Means Variation <https://elki-project.github.io/tutorial/same-size_k_means>`_.
    """
    # distances
    distances = euclidean_distances(
        centers, X, Y_norm_squared=x_squared_norms, squared=True)
    distances = distances.T

    if strategy == 'gain_p':
        labels[:] = numpy.argmin(distances, axis=1)
    else:
        # We assume labels comes from a previous iteration.
        pass

    strategy_coef = _compute_strategy_coefficient(distances, strategy, labels)
    distance_linear = linearize_matrix(distances, strategy_coef)
    sorted_distances = distance_linear[distance_linear[:, 3].argsort()]
    distances_close[:] = 0

    # counters
    ave = limit
    counters[:] = 0
    for i in labels:
        counters[i] += 1
    leftclose[:] = counters[:] - ave
    leftclose[leftclose < 0] = 0
    nover = X.shape[0] - ave * counters.shape[0]
    sumi = nover - leftclose.sum()
    if sumi != 0:
        if state is None:
            state = numpy.random.RandomState()  # pylint: disable=E1101

        def loopf(h, sumi):
            if sumi < 0 and leftclose[h] > 0:  # pylint: disable=R1716
                sumi -= leftclose[h]
                leftclose[h] = 0
            elif sumi > 0 and leftclose[h] == 0:
                leftclose[h] = 1
                sumi += 1
            return sumi

        it = 0
        while sumi != 0:
            h = state.randint(0, counters.shape[0])
            sumi = loopf(h, sumi)
            it += 1
            if it > counters.shape[0] * 2:
                break
        for h in range(counters.shape[0]):
            if sumi == 0:
                break
            sumi = loopf(h, sumi)

    transfer = {}

    for i in range(0, sorted_distances.shape[0]):
        gain = sorted_distances[i, 3]
        ind = int(sorted_distances[i, 1])
        dest = int(sorted_distances[i, 2])
        cur = labels[ind]
        if distances_close[ind]:
            continue
        if cur == dest:
            continue
        if ((counters[dest] < ave + leftclose[dest]) and
                (counters[cur] > ave + leftclose[cur])):
            labels[ind] = dest
            counters[cur] -= 1
            counters[dest] += 1
            distances_close[ind] = 1  # moved
        else:
            cp = transfer.get((dest, cur), [])
            while len(cp) > 0:
                g, destind = cp[0]
                if distances_close[destind]:
                    del cp[0]
                else:
                    break
            if len(cp) > 0:
                g, destind = cp[0]
                if g + gain < 0:
                    del cp[0]
                    labels[ind] = dest
                    labels[destind] = cur
                    add = False
                    distances_close[ind] = 1  # moved
                    distances_close[destind] = 1  # moved
                else:
                    add = True
            else:
                add = True
            if add:
                # We add the point to the list of points willing to transfer.
                if (cur, dest) not in transfer:
                    transfer[cur, dest] = []
                gain = sorted_distances[i, 3]
                bisect.insort(transfer[cur, dest], (gain, ind))

    neg = (counters < ave).sum()
    if neg > 0:
        raise RuntimeError(  # pragma: no cover
            "The algorithm failed, counters={0}".format(counters))

    _switch_clusters(labels, distances)
    distances_close[:] = distances[numpy.arange(X.shape[0]), labels]

    return distances


def _constraint_kmeans_weights(X, labels, sample_weight, centers, inertia, it,
                               max_iter, verbose=0, state=None, learning_rate=1.,
                               history=False, fLOG=None):
    """
    Runs KMeans iterator but weights cluster among them.

    @param      X               features
    @param      labels          initialized labels (unused)
    @param      sample_weight   sample weight
    @param      centers         initialized centers
    @param      inertia         initialized inertia (unused)
    @param      it              number of iteration already done
    @param      max_iter        maximum of number of iteration
    @param      verbose         verbose
    @param      state           random state
    @param      learning_rate   learning rate
    @param      history         keeps all centers accross iterations
    @param      fLOG            logging function (needs to be specified otherwise
                                verbose has no effects)
    @return                     tuple (best_labels, best_centers, best_inertia, weights, it)
    """
    if isinstance(X, DataFrame):
        X = X.values
    n_clusters = centers.shape[0]
    best_inertia = None
    best_iter = None
    weights = numpy.ones(centers.shape[0])
    if sample_weight is None:
        sw = numpy.ones((X.shape[0],))
    else:
        sw = sample_weight

    if scipy.sparse.issparse(X):
        _centers_fct = _centers_sparse
    else:
        _centers_fct = _centers_dense

    total_inertia = _inertia(X, sw)
    all_centers = []

    while it < max_iter:

        # compute new clusters
        centers = _centers_fct(
            X, sw, labels, n_clusters, None)
        if history:
            all_centers.append(centers)

        # association
        labels = _constraint_association_weights(X, centers, sw, weights)
        if len(set(labels)) != centers.shape[0]:
            if verbose and fLOG:  # pragma: no cover
                if isinstance(verbose, int) and verbose >= 10:
                    fLOG("CKMeans new weights: w=%r" % weights)
                else:
                    fLOG("CKMeans new weights")
            weights[:] = 1
            labels = _constraint_association_weights(X, centers, sw, weights)

        # inertia
        inertia, diff = _labels_inertia_weights(
            X, centers, sw, weights, labels, total_inertia)
        if numpy.isnan(inertia):
            raise RuntimeError(  # pragma: no cover
                "nanNobs={} Nclus={}\ninertia={}\nweights={}\ndiff={}\nlabels={}".format(
                    X.shape[0], centers.shape[0], inertia, weights, diff,
                    set(labels)))

        # best option so far?
        if best_inertia is None or inertia < best_inertia:
            best_inertia = inertia
            best_centers = centers.copy()
            best_labels = labels.copy()
            best_weights = weights.copy()
            best_iter = it

        # moves weights
        weights, _ = _adjust_weights(X, sw, weights, labels,
                                     learning_rate / (it + 10))

        it += 1
        if verbose and fLOG:
            if isinstance(verbose, int) and verbose >= 10:
                fLOG("CKMeans %d/%d inertia=%f (%f T=%f) dw=%r w=%r" % (
                    it, max_iter, inertia, best_inertia, total_inertia,
                    diff, weights))
            elif isinstance(verbose, int) and verbose >= 5:
                hist = Counter(labels)
                fLOG("CKMeans %d/%d inertia=%f (%f) hist=%r" % (
                    it, max_iter, inertia, best_inertia, hist))
            else:
                fLOG("CKMeans %d/%d inertia=%f (%f T=%f)" % (  # pragma: no cover
                    it, max_iter, inertia, best_inertia, total_inertia))

        # early stop
        if (best_inertia is not None and inertia >= best_inertia and
                it > best_iter + 5 and numpy.abs(diff).sum() <= weights.shape[0] / 2):
            break

    return (best_labels, best_centers, best_inertia, best_weights,
            it, all_centers)


def _constraint_association_weights(X, centers, sw, weights):
    """
    Associates points to clusters.

    @param      X           features
    @param      centers     centers
    @param      sw          sample weights
    @param      weights     cluster weights
    @return                 labels
    """
    dist = cdist(X, centers) * weights.reshape((1, -1))
    index = numpy.argmin(dist, axis=1)
    return index


def _inertia(X, sw):
    """
    Computes total weighted inertia.

    @param      X               features
    @param      sw              sample weights
    @return                     inertia
    """
    bary = numpy.mean(X, axis=0)
    diff = X - bary
    norm = numpy.linalg.norm(diff, axis=1)
    if sw is not None:
        norm *= sw
    return sw.sum()


def _labels_inertia_weights(X, centers, sw, weights, labels, total_inertia):
    """
    Computes weighted inertia. It also adds a fraction
    of the whole inertia depending on how balanced the
    clusters are.

    @param      X               features
    @param      centers         centers
    @param      sw              sample weights
    @param      weights         cluster weights
    @param      labels          labels
    @param      total_inertia   total inertia
    @return                     inertia
    """
    www, exp, _ = _compute_balance(X, sw, labels, centers.shape[0])
    www -= exp
    wwwa = numpy.square(www)
    ratio = wwwa.sum() ** 0.5 / X.shape[0]
    dist = cdist(X, centers) * weights.reshape((1, -1)) * sw.reshape((-1, 1))
    return dist.sum() + ratio * total_inertia, www


def _compute_balance(X, sw, labels, nbc=None):
    """
    Computes weights difference.

    @param      X           features
    @param      sw          sample weights
    @param      labels      known labels
    @param      nbc         number of clusters
    @return                 (weights per cluster, expected weight,
                            total weight)
    """
    if nbc is None:
        nbc = labels.max() + 1
    N = numpy.float64(nbc)
    www = numpy.zeros((nbc, ), dtype=numpy.float64)
    if sw is None:
        for la in labels:
            www[la] += 1
    else:
        for w, la in zip(sw, labels):
            www[la] += w
    exp = www.sum() / N
    return www, exp, N


def _adjust_weights(X, sw, weights, labels, lr):
    """
    Changes *weights* mapped to every cluster.
    *weights < 1* are used for big clusters,
    *weights > 1* are used for small clusters.

    @param      X           features
    @param      centers     centers
    @param      sw          sample weights
    @param      weights     cluster weights
    @param      lr          learning rate
    @param      labels      known labels
    @return                 labels
    """
    www, exp, N = _compute_balance(X, sw, labels, weights.shape[0])

    for i in range(0, weights.shape[0]):
        nw = (www[i] - exp) / exp
        delta = nw * lr
        weights[i] += delta
        N += delta

    return weights / N * weights.shape[0], www



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a *transform* which follows the smae API
as every :epkg:`scikit-learn` transform.
"""
from .sklearn_base import SkBase


class SkBaseTransform(SkBase):
    """
    Pattern of a *learner* which follows the same API que :epkg:`scikit-learn`.
    """

    def __init__(self, **kwargs):
        """
        Stores the parameters.
        """
        SkBase.__init__(self, **kwargs)

    ###################
    # API scikit-learn
    ###################

    def fit(self, X, y=None, **kwargs):
        """
        Trains a model.

        @param      X               features
        @param      y               targets
        @return                     self
        """
        raise NotImplementedError()

    def transform(self, X):
        """
        Transforms the data.

        @param      X   features
        @return         predictions
        """
        raise NotImplementedError()

    def fit_transform(self, X, y=None, **kwargs):
        """
        Trains and transforms the data.

        @param      X               features
        @param      y               targets
        @return                     self
        """
        self.fit(X, y=y, **kwargs)
        return self.transform(X)



"""
@file
@brief Timeseries plots.
"""
import calendar
import datetime
import matplotlib.pyplot as plt
import matplotlib.patches as patches


def plot_week_timeseries(time, value, normalise=True,
                         label=None, h=0.85, value2=None,
                         label2=None, daynames=None,
                         xfmt="%1.0f", ax=None):
    """
    Shows a timeseries dispatched by days as bars.

    @param      time        dates
    @param      value       values to display as bars.
    @param      normalise   normalise data before showing it
    @param      label       label of the series
    @param      values2     second series to show as a line
    @param      label2      label of the second series
    @param      daynames    names to use for week day names (default is English)
    @param      xfmt        format number of the X axis
    @param      ax          existing axis
    @return                 axis

    .. plot::

        import datetime
        import matplotlib.pyplot as plt
        from mlinsights.timeseries.datasets import artificial_data
        from mlinsights.timeseries.agg import aggregate_timeseries
        from mlinsights.timeseries.plotting import plot_week_timeseries

        dt1 = datetime.datetime(2019, 8, 1)
        dt2 = datetime.datetime(2019, 9, 1)
        data = artificial_data(dt1, dt2, minutes=15)
        print(data.head())

        agg = aggregate_timeseries(data, per='week')
        plot_week_timeseries(
            agg['weektime'], agg['y'], label="y",
            value2=agg['y']/2, label2="y/2", normalise=False)
        plt.show()
    """
    if time.shape[0] != value.shape[0]:
        raise AssertionError("Dimension mismatch")  # pragma: no cover

    def coor(ti):
        days = ti.days
        x = days
        y = ti.seconds
        return x, y

    max_value = value.max()
    if value2 is not None:
        max_value = max(max_value, value2.max())
        value2 = value2 / max_value
    value = value / max_value
    input_maxy = 1.

    if ax is None:
        ax = plt.gca()

    # bars
    delta = None
    maxx, maxy = None, None
    first = True
    for i in range(time.shape[0]):
        ti = time[i]
        if i < time.shape[0] - 1:
            ti1 = time[i + 1]
            delta = (ti1 - ti) if delta is None else min(delta, ti1 - ti)
            if delta == 0:
                raise RuntimeError(  # pragma: no cover
                    "The timeseries contains duplicated time values.")
        else:
            ti1 = ti + delta
        x1, y1 = coor(ti)
        x2, y2 = coor(ti1)
        if y2 < y1:
            x2, y2 = coor(ti + delta)
        y2 = y1 + (y2 - y1) * h
        if first and label:
            ax.plot([x1, x1 + value[i] * 0.8], [y1, y1],
                    'b', alpha=0.5, label=label)
            first = False
        if maxx is None:
            maxx = (x1, x1 + input_maxy)
            maxy = (y1, y2)
        else:
            maxx = (min(x1, maxx[0]), max(x1 + input_maxy, maxx[1]))
            maxy = (min(y1, maxy[0]), max(y2, maxy[1]))

        rect = patches.Rectangle((x1, y1), value[i] * h, y2 - y1,
                                 linewidth=1, edgecolor=None,
                                 facecolor='b', fill=True,
                                 alpha=0.5)

        ax.add_patch(rect)

    # days border
    xticks = []
    if daynames is None:
        daynames = list(calendar.day_name)

    maxx = [(maxx[0] // 7) * 7, maxx[1]]
    new_ymin = maxy[0] - (maxy[1] * 0.025 + maxy[0] * 0.975 - maxy[0])
    for i in range(int(maxx[0]), int(maxx[1] + 0.1)):
        x1i = maxx[0] + input_maxy * i
        x2i = x1i + input_maxy
        xticks.append(x1i)
        ax.plot([x1i, x1i + input_maxy], [new_ymin, new_ymin], 'k', alpha=0.5)
        ax.plot([x1i, x1i + input_maxy], [maxy[1], maxy[1]], 'k', alpha=0.5)
        ax.plot([x1i, x1i], [maxy[0], maxy[1]], 'k', alpha=0.5)
        ax.plot([x2i, x2i], [maxy[0], maxy[1]], 'k', alpha=0.5)
        ax.text(x1i, new_ymin, daynames[i])

    # invert y axis
    ax.invert_yaxis()

    # change y labels
    nby = len(ax.get_yticklabels())
    ys = ax.get_yticks()
    ylabels = []
    for i in range(nby):
        dh = ys[i]
        dt = datetime.timedelta(seconds=dh)
        tx = "%dh%02d" % (dt.seconds // 3600,
                          60 * (dt.seconds / 3600 - dt.seconds // 3600))
        ylabels.append(tx)
    ax.set_yticklabels(ylabels)

    # change x labels
    xs = ax.get_xticks()
    xticks = []
    xlabels = []
    for i in range(0, len(xs) - 1):
        if xs[i] < 0:
            continue
        dx = xs[i] - int(xs[i] / input_maxy) * input_maxy
        xlabels.append(dx if normalise else (dx * max_value))
        xticks.append(xs[i])
        dx = (xs[i] + xs[i + 1]) / 2
        dx = dx - int(dx / input_maxy) * input_maxy
        xlabels.append(dx if normalise else (dx * max_value))
        xticks.append((xs[i] + xs[i + 1]) / 2)
    if len(xticks) < len(xlabels):
        xticks.append(xs[-1])
    ax.set_xticks(xticks)
    ax.set_xticklabels(
        [xfmt % x for x in xlabels] if xfmt else xlabels)

    ax.tick_params(axis='x', rotation=30)

    # value2
    if value2 is not None:
        value = value2.copy()
        if normalise:
            value = value / max_value

        first = True
        xs = []
        ys = []
        for i in range(time.shape[0]):
            ti = time[i]
            if i < time.shape[0] - 1:
                ti1 = time[i + 1]
            else:
                ti1 = ti + delta
            x1, y1 = coor(ti)
            x2, y2 = coor(ti1)
            if y2 < y1:
                x2, y2 = coor(ti + delta)
            y2 = y1 + (y2 - y1) * h

            x2 = x1 + value[i] * h

            if len(ys) > 0 and y2 < ys[-1]:
                if first and label2 is not None:
                    ax.plot(xs, ys, color='orange', linewidth=2, label=label2)
                    first = False
                else:
                    ax.plot(xs, ys, color='orange', linewidth=2)
                xs, ys = [], []

            xs.append(x2)
            ys.append((y1 + y2) / 2)

        if len(xs) > 0:
            ax.plot(xs, ys, color='orange', linewidth=2)

    return ax



"""
@file
@brief Implements a piecewise linear regression.
"""
import numpy
import numpy.random
from sklearn.base import RegressorMixin, clone, BaseEstimator
from sklearn.utils._joblib import Parallel, delayed
from sklearn.utils.fixes import _joblib_parallel_args
try:
    from tqdm import tqdm
except ImportError:  # pragma: no cover
    pass


class IntervalRegressor(BaseEstimator, RegressorMixin):
    """
    Trains multiple regressors to provide a confidence
    interval on prediction. It only works for
    single regression. Every training is made with a new
    sample of the training data, parameter *alpha*
    let the user choose the size of this sample.
    A smaller *alpha* increases the variance
    of the predictions. The current implementation
    draws sample by random but keeps the weight associated
    to each of them. Another way could be to draw
    a weighted sample but give them uniform weights.
    """

    def __init__(self, estimator=None, n_estimators=10, n_jobs=None,
                 alpha=1., verbose=False):
        """
        @param      estimator           predictor trained on every bucket
        @param      n_estimators        number of estimators to train
        @param      n_jobs              number of parallel jobs (for training and predicting)
        @param      alpha               proportion of samples resampled for each training
        @param      verbose             boolean or use ``'tqdm'`` to use :epkg:`tqdm`
                                        to fit the estimators
        """
        BaseEstimator.__init__(self)
        RegressorMixin.__init__(self)
        if estimator is None:
            raise ValueError("estimator cannot be null.")  # pragma: no cover
        self.estimator = estimator
        self.n_jobs = n_jobs
        self.alpha = alpha
        self.verbose = verbose
        self.n_estimators = n_estimators

    @property
    def n_estimators_(self):
        """
        Returns the number of estimators = the number of buckets
        the data was split in.
        """
        return len(self.estimators_)

    def fit(self, X, y, sample_weight=None):
        """
        Trains the binner and an estimator on every
        bucket.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :param y: target
        :param sample_weight: sample weights
        :return: self: returns an instance of self.

        Fitted attributes:

        * `binner_`: binner
        * `estimators_`: dictionary of estimators, each of them
            mapped to a leave to the tree
        * `mean_estimator_`: estimator trained on the whole
            datasets in case the binner can find a bucket for
            a new observation
        * `dim_`: dimension of the output
        * `mean_`: average targets
        """
        self.estimators_ = []
        estimators = [clone(self.estimator) for i in range(self.n_estimators)]

        loop = tqdm(range(len(estimators))
                    ) if self.verbose == 'tqdm' else range(len(estimators))
        verbose = 1 if self.verbose == 'tqdm' else (1 if self.verbose else 0)

        def _fit_piecewise_estimator(i, est, X, y, sample_weight, alpha):
            new_size = int(X.shape[0] * alpha + 0.5)
            rnd = numpy.random.randint(0, X.shape[0] - 1, new_size)
            Xr = X[rnd]
            yr = y[rnd]
            sr = sample_weight[rnd] if sample_weight else None
            return est.fit(Xr, yr, sr)

        self.estimators_ = \
            Parallel(n_jobs=self.n_jobs, verbose=verbose,
                     **_joblib_parallel_args(prefer='threads'))(
                delayed(_fit_piecewise_estimator)(
                    i, estimators[i], X, y, sample_weight, self.alpha)
                for i in loop)

        return self

    def predict_all(self, X):
        """
        Computes the predictions for all estimators.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions
        """
        container = numpy.empty((X.shape[0], len(self.estimators_)))
        for i, est in enumerate(self.estimators_):
            pred = est.predict(X)
            container[:, i] = pred
        return container

    def predict(self, X):
        """
        Computes the average predictions.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions
        """
        preds = self.predict_all(X)
        return preds.mean(axis=1)

    def predict_sorted(self, X):
        """
        Computes the predictions for all estimators.
        Sorts them for all observations.

        :param X: features, *X* is converted into an array if *X* is a dataframe
        :return: predictions sorted for each observation
        """
        preds = self.predict_all(X)
        for i in range(preds.shape[0]):
            preds[i, :] = numpy.sort(preds[i, :])
        return preds



"""
@file
@brief Dig into pipelines.
"""
import textwrap
import warnings
from types import MethodType
from sklearn.base import TransformerMixin, ClassifierMixin, RegressorMixin, BaseEstimator
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.compose import ColumnTransformer, TransformedTargetRegressor


def enumerate_pipeline_models(pipe, coor=None, vs=None):
    """
    Enumerates all the models within a pipeline.

    @param      pipe        *scikit-learn* pipeline
    @param      coor        current coordinate
    @param      vs          subset of variables for the model, None for all
    @return                 iterator on models ``tuple(coordinate, model)``

    See notebook :ref:`visualizepipelinerst`.
    """
    if coor is None:
        coor = (0,)
    if pipe == "passthrough":
        class PassThrough:
            "dummy class to help display"
            pass
        yield coor, PassThrough(), vs
    else:
        yield coor, pipe, vs
        if hasattr(pipe, 'transformer_and_mapper_list') and len(pipe.transformer_and_mapper_list):
            # azureml DataTransformer
            raise NotImplementedError(  # pragma: no cover
                "Unable to handle this specific case.")
        elif hasattr(pipe, 'mapper') and pipe.mapper:
            # azureml DataTransformer
            for couple in enumerate_pipeline_models(pipe.mapper, coor + (0,)):
                yield couple
        elif hasattr(pipe, 'built_features'):  # pragma: no cover
            # sklearn_pandas.dataframe_mapper.DataFrameMapper
            for i, (columns, transformers, _) in enumerate(pipe.built_features):
                if isinstance(columns, str):
                    columns = (columns,)
                if transformers is None:
                    yield (coor + (i,)), None, columns
                else:
                    for couple in enumerate_pipeline_models(transformers, coor + (i,), columns):
                        yield couple
        elif isinstance(pipe, Pipeline):
            for i, (_, model) in enumerate(pipe.steps):
                for couple in enumerate_pipeline_models(model, coor + (i,)):
                    yield couple
        elif isinstance(pipe, ColumnTransformer):
            for i, (_, fitted_transformer, column) in enumerate(pipe.transformers):
                for couple in enumerate_pipeline_models(
                        fitted_transformer, coor + (i,), column):
                    yield couple
        elif isinstance(pipe, FeatureUnion):
            for i, (_, model) in enumerate(pipe.transformer_list):
                for couple in enumerate_pipeline_models(model, coor + (i,)):
                    yield couple
        elif isinstance(pipe, TransformedTargetRegressor):
            raise NotImplementedError(  # pragma: no cover
                "Not yet implemented for TransformedTargetRegressor.")
        elif isinstance(pipe, (TransformerMixin, ClassifierMixin, RegressorMixin)):
            pass
        elif isinstance(pipe, BaseEstimator):  # pragma: no cover
            pass
        else:
            raise TypeError(  # pragma: no cover
                "pipe is not a scikit-learn object: {}\n{}".format(type(pipe), pipe))


class BaseEstimatorDebugInformation:
    """
    Stores information when the outputs of a pipeline
    is computed. It as added by function
    @see fct alter_pipeline_for_debugging.
    """

    def __init__(self, model):
        self.model = model
        self.inputs = {}
        self.outputs = {}
        self.methods = {}
        if hasattr(model, "transform") and callable(model.transform):
            model._debug_transform = model.transform
            self.methods["transform"] = lambda model, X: model._debug_transform(
                X)
        if hasattr(model, "predict") and callable(model.predict):
            model._debug_predict = model.predict
            self.methods["predict"] = lambda model, X: model._debug_predict(X)
        if hasattr(model, "predict_proba") and callable(model.predict_proba):
            model._debug_predict_proba = model.predict_proba
            self.methods["predict_proba"] = lambda model, X: model._debug_predict_proba(
                X)
        if hasattr(model, "decision_function") and callable(model.decision_function):
            model._debug_decision_function = model.decision_function
            self.methods["decision_function"] = lambda model, X: model._debug_decision_function(
                X)

    def __repr__(self):
        """
        usual
        """
        return self.to_str()

    def to_str(self, nrows=5):
        """
        Tries to produce a readable message.
        """
        rows = ['BaseEstimatorDebugInformation({})'.format(
            self.model.__class__.__name__)]
        for k in sorted(self.inputs):
            if k in self.outputs:
                rows.append('  ' + k + '(')
                self.display(self.inputs[k], nrows)
                rows.append(textwrap.indent(
                    self.display(self.inputs[k], nrows), '   '))
                rows.append('  ) -> (')
                rows.append(textwrap.indent(
                    self.display(self.outputs[k], nrows), '   '))
                rows.append('  )')
            else:
                raise KeyError(  # pragma: no cover
                    "Unable to find output for method '{}'.".format(k))
        return "\n".join(rows)

    def display(self, data, nrows):
        """
        Displays the first
        """
        text = str(data)
        rows = text.split('\n')
        if len(rows) > nrows:
            rows = rows[:nrows]
            rows.append('...')
        if hasattr(data, 'shape'):
            rows.insert(0, "shape={} type={}".format(data.shape, type(data)))
        else:
            rows.insert(0, "type={}".format(type(data)))
        return "\n".join(rows)


def alter_pipeline_for_debugging(pipe):
    """
    Overwrite methods *transform*, *predict*, *predict_proba*
    or *decision_function* to collect the last inputs and outputs
    seen in these methods.

    @param      pipe        *scikit-learn* pipeline

    The object *pipe* is modified, it should be copied
    before calling this function if you need the object
    untouched after that. The prediction is slower.
    See notebook :ref:`visualizepipelinerst`.
    """

    def transform(self, X, *args, **kwargs):
        self._debug.inputs['transform'] = X
        y = self._debug.methods['transform'](self, X, *args, **kwargs)
        self._debug.outputs['transform'] = y
        return y

    def predict(self, X, *args, **kwargs):
        self._debug.inputs['predict'] = X
        y = self._debug.methods['predict'](self, X, *args, **kwargs)
        self._debug.outputs['predict'] = y
        return y

    def predict_proba(self, X, *args, **kwargs):
        self._debug.inputs['predict_proba'] = X
        y = self._debug.methods['predict_proba'](self, X, *args, **kwargs)
        self._debug.outputs['predict_proba'] = y
        return y

    def decision_function(self, X, *args, **kwargs):
        self._debug.inputs['decision_function'] = X
        y = self._debug.methods['decision_function'](self, X, *args, **kwargs)
        self._debug.outputs['decision_function'] = y
        return y

    new_methods = {
        'decision_function': decision_function,
        'transform': transform,
        'predict': predict,
        'predict_proba': predict_proba,
    }

    if hasattr(pipe, '_debug'):
        raise RuntimeError(  # pragma: no cover
            "The same operator cannot be used twice in "
            "the same pipeline or this method was called "
            "a second time.")

    for model_ in enumerate_pipeline_models(pipe):
        model = model_[1]
        model._debug = BaseEstimatorDebugInformation(model)
        for k in model._debug.methods:
            try:
                setattr(model, k, MethodType(new_methods[k], model))
            except AttributeError:  # pragma: no cover
                warnings.warn("Unable to overwrite method '{}' for class "
                              "{}.".format(k, type(model)))



"""
@file
@brief Featurizers for machine learned models.
"""
import numpy
from sklearn.base import BaseEstimator, RegressorMixin, MultiOutputMixin
from sklearn.decomposition import NMF, TruncatedSVD


class ApproximateNMFPredictor(BaseEstimator, RegressorMixin, MultiOutputMixin):
    """
    Converts :epkg:`sklearn:decomposition:NMF` into
    a predictor so that the prediction does not involve
    training even for new observations. The class uses a
    :epkg:`sklearn:decomposition:TruncatedSVD` of the components
    found by the :epkg:`sklearn:decomposition:NMF`.
    The prediction projects the test data into
    the components vector space and retrieves them back
    into their original space. The issue is it does not
    necessarily produce results with only positive
    results as the :epkg:`sklearn:decomposition:NMF`
    would do unless parameter *force_positive* is True.

    .. runpython::
        :showcode:

        import numpy
        from mlinsights.mlmodel.anmf_predictor import ApproximateNMFPredictor

        train = numpy.array([[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0],
                           [1, 0, 0, 0], [1, 0, 0, 0]], dtype=numpy.float64)
        train[:train.shape[1], :] += numpy.identity(train.shape[1])

        model = ApproximateNMFPredictor(n_components=2,
                                        force_positive=True)
        model .fit(train)

        test = numpy.array([[1, 1, 1, 0]], dtype=numpy.float64)
        pred = model.predict(test)
        print(pred)
    """

    def __init__(self, force_positive=False, **kwargs):
        """
        *kwargs* should contains parameters
        for :epkg:`sklearn:decomposition:NMF`.
        The parameter *force_positive* removes all
        negative predictions and replaces by zero.
        """
        BaseEstimator.__init__(self)
        RegressorMixin.__init__(self)
        MultiOutputMixin.__init__(self)
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.force_positive = force_positive

    @classmethod
    def _get_param_names(cls):
        """
        Returns the list of parameters
        of the estimator.
        """
        res = NMF._get_param_names()
        res = res + ["force_positive"]
        return res

    def get_params(self, deep=True):
        """
        Returns the parameters of the estimator
        as a dictionary.
        """
        res = {}
        for k in self.__class__._get_param_names():
            if hasattr(self, k):
                res[k] = getattr(self, k)
        return res

    def fit(self, X, y=None):
        """
        Trains a :epkg:`sklearn:decomposition:NMF`
        then a multi-output regressor.
        """
        params = self.get_params()
        if 'force_positive' in params:
            del params['force_positive']
        self.estimator_nmf_ = NMF(**params)
        self.estimator_nmf_.fit(X)
        self.estimator_svd_ = TruncatedSVD(
            n_components=self.estimator_nmf_.n_components_)
        self.estimator_svd_.fit(self.estimator_nmf_.components_)
        return self

    def predict(self, X):
        """
        Predicts based on the multi-output regressor.
        The output has the same dimension as *X*.
        """
        proj = self.estimator_svd_.transform(X)
        pred = self.estimator_svd_.inverse_transform(proj)
        if self.force_positive:
            zeros = numpy.zeros(
                (1, pred.shape[1]), dtype=pred.dtype)  # pylint: disable=E1101,E1136
            pred = numpy.maximum(pred, zeros)  # pylint: disable=E1111
        return pred



"""
@file
@brief Implements a slightly different
version of the :epkg:`sklearn:compose:TransformedTargetRegressor`.
"""
from sklearn.base import BaseEstimator, RegressorMixin, ClassifierMixin, clone
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import r2_score, accuracy_score
from .sklearn_transform_inv import BaseReciprocalTransformer
from .sklearn_transform_inv_fct import FunctionReciprocalTransformer, PermutationReciprocalTransformer


def _common_get_transform(transformer, is_regression):
    if isinstance(transformer, str):
        closest = is_regression
        if transformer == 'permute':
            return PermutationReciprocalTransformer(closest=closest)
        else:
            return FunctionReciprocalTransformer(transformer)
    elif isinstance(transformer, BaseReciprocalTransformer):
        return clone(transformer)
    raise TypeError("Transformer {} must be a string or on object of type "
                    "BaseReciprocalTransformer.".format(type(transformer)))


class TransformedTargetRegressor2(BaseEstimator, RegressorMixin):
    """
    Meta-estimator to regress on a transformed target.
    Useful for applying a non-linear transformation in regression
    problems.

    Parameters
    ----------
    regressor : object, default=LinearRegression()
        Regressor object such as derived from ``RegressorMixin``. This
        regressor will automatically be cloned each time prior to fitting.
    transformer : str or object of type @see cl BaseReciprocalTransformer

    Attributes
    ----------
    regressor_ : object
        Fitted regressor.
    transformer_ : object
        Transformer used in ``fit`` and ``predict``.

    Examples
    --------

    .. runpython::
        :showcode:

        import numpy
        from sklearn.linear_model import LinearRegression
        from mlinsights.mlmodel import TransformedTargetRegressor2

        tt = TransformedTargetRegressor2(regressor=LinearRegression(),
                                         transformer='log')
        X = numpy.arange(4).reshape(-1, 1)
        y = numpy.exp(2 * X).ravel()
        print(tt.fit(X, y))
        print(tt.score(X, y))
        print(tt.regressor_.coef_)

    See notebook :ref:`sklearntransformedtargetrst` for a more complete example.
    """

    def __init__(self, regressor=None, transformer=None):
        self.regressor = regressor
        self.transformer = transformer

    def fit(self, X, y, sample_weight=None):
        """
        Fits the model according to the given training data.

        :param X: {array-like, sparse matrix}, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
        :param y: array-like, shape (n_samples,)
            Target values.
        :param sample_weight: array-like, shape (n_samples,) optional
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.
        :return: self, object
        """
        self.transformer_ = _common_get_transform(self.transformer, True)
        self.transformer_.fit(X, y, sample_weight=sample_weight)
        X_trans, y_trans = self.transformer_.transform(X, y)

        if self.regressor is None:
            self.regressor_ = LinearRegression()
        else:
            self.regressor_ = clone(self.regressor)

        if sample_weight is None:
            self.regressor_.fit(X_trans, y_trans)
        else:
            self.regressor_.fit(X_trans, y_trans, sample_weight=sample_weight)

        return self

    def predict(self, X):
        """
        Predicts using the base regressor, applying inverse.

        :param X: {array-like, sparse matrix}, shape = (n_samples, n_features)
            Samples.
        :return: y_hat : array, shape = (n_samples,)
            Predicted values.
        """
        if not hasattr(self, 'regressor_'):
            raise NotFittedError(
                "This instance {} is not fitted yet. Call 'fit' with "
                "appropriate arguments before using this method.".format(
                    type(self)))
        X_trans, _ = self.transformer_.transform(X, None)
        pred = self.regressor_.predict(X_trans)

        inv = self.transformer_.get_fct_inv()
        _, pred_inv = inv.transform(X_trans, pred)
        return pred_inv

    def score(self, X, y, sample_weight=None):
        """
        Scores the model with
        :epkg:`sklearn:metrics:r2_score`.
        """
        yp = self.predict(X)
        return r2_score(y, yp, sample_weight=sample_weight)

    def _more_tags(self):
        return {'poor_score': True, 'no_validation': True}


class TransformedTargetClassifier2(BaseEstimator, ClassifierMixin):
    """
    Meta-estimator to classify on a transformed target.
    Useful for applying permutation transformation in classification
    problems.

    Parameters
    ----------
    classifier : object, default=LogisticRegression()
        Classifier object such as derived from ``ClassifierMixin``. This
        classifier will automatically be cloned each time prior to fitting.
    transformer : str or object of type @see cl BaseReciprocalTransformer

    Attributes
    ----------
    classifier_ : object
        Fitted classifier.
    transformer_ : object
        Transformer used in ``fit``, ``predict``, ``decision_function``,
        ``predict_proba``.

    Examples
    --------

    .. runpython::
        :showcode:

        import numpy
        from sklearn.linear_model import LogisticRegression
        from mlinsights.mlmodel import TransformedTargetClassifier2

        tt = TransformedTargetClassifier2(classifier=LogisticRegression(),
                                         transformer='permute')
        X = numpy.arange(4).reshape(-1, 1)
        y = numpy.array([0, 1, 0, 1])
        print(tt.fit(X, y))
        print(tt.score(X, y))
        print(tt.classifier_.coef_)

    See notebook :ref:`sklearntransformedtargetrst` for a more complete example.
    """

    def __init__(self, classifier=None, transformer=None):
        self.classifier = classifier
        self.transformer = transformer

    def fit(self, X, y, sample_weight=None):
        """
        Fits the model according to the given training data.

        :param X: {array-like, sparse matrix}, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
        :param y: array-like, shape (n_samples,)
            Target values.
        :param sample_weight: array-like, shape (n_samples,) optional
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.
        :return: self, object
        """
        self.transformer_ = _common_get_transform(self.transformer, False)
        self.transformer_.fit(X, y, sample_weight=sample_weight)
        X_trans, y_trans = self.transformer_.transform(X, y)

        if self.classifier is None:
            self.classifier_ = LogisticRegression()
        else:
            self.classifier_ = clone(self.classifier)

        if sample_weight is None:
            self.classifier_.fit(X_trans, y_trans)
        else:
            self.classifier_.fit(X_trans, y_trans, sample_weight=sample_weight)

        return self

    def _check_is_fitted(self):
        if not hasattr(self, 'classifier_'):
            raise NotFittedError(
                "This instance {} is not fitted yet. Call 'fit' with "
                "appropriate arguments before using this method.".format(
                    type(self)))

    @property
    def classes_(self):
        """
        Returns the classes.
        """
        self._check_is_fitted()
        inv = self.transformer_.get_fct_inv()
        _, pred_inv = inv.transform(None, self.classifier_.classes_)
        return pred_inv

    def _apply(self, X, method):
        """
        Calls *predict*, *predict_proba* or *decision_function*
        using the base classifier, applying inverse.

        :param X: {array-like, sparse matrix}, shape = (n_samples, n_features)
            Samples.
        :return: y_hat, array, shape = (n_samples,)
            Predicted values.
        """
        self._check_is_fitted()
        if not hasattr(self.classifier_, method):
            raise RuntimeError("Unable to find method '{}' in model {}.".format(
                method, type(self.classifier_)))
        meth = getattr(self.classifier_, method)
        X_trans, _ = self.transformer_.transform(X, None)
        pred = meth(X_trans)
        inv = self.transformer_.get_fct_inv()
        _, pred_inv = inv.transform(X_trans, pred)
        return pred_inv

    def predict(self, X):
        """
        Predicts using the base classifier, applying inverse.

        :param X: {array-like, sparse matrix}, shape = (n_samples, n_features)
            Samples.
        :return: y_hat, array, shape = (n_samples,)
            Predicted values.
        """
        return self._apply(X, 'predict')

    def predict_proba(self, X):
        """
        Predicts using the base classifier, applying inverse.

        :param X: {array-like, sparse matrix}, shape = (n_samples, n_features)
            Samples.
        :return: predict probabilities, array, shape = (n_samples, n_classes)
            Predicted values.
        """
        return self._apply(X, 'predict_proba')

    def decision_function(self, X):
        """
        Predicts using the base classifier, applying inverse.

        :param X: {array-like, sparse matrix}, shape = (n_samples, n_features)
            Samples.
        :return: raw score : array, shape = (n_samples, ?)
        """
        return self._apply(X, 'decision_function')

    def score(self, X, y, sample_weight=None):
        """
        Scores the model with
        :epkg:`sklearn:metrics:accuracy_score`.
        """
        yp = self.predict(X)
        return accuracy_score(y, yp, sample_weight=sample_weight)

    def _more_tags(self):
        return {'poor_score': True, 'no_validation': True}



"""
@file
@brief Implements a transformation which can be put in a pipeline to transform categories in
integers.
"""
import numpy
import pandas
from sklearn.base import BaseEstimator, TransformerMixin


class CategoriesToIntegers(BaseEstimator, TransformerMixin):
    """
    Does something similar to what
    `DictVectorizer <http://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.DictVectorizer.html>`_
    does but in a transformer. The method *fit* retains all categories,
    the method *transform* transforms categories into integers.
    Categories are sorted by columns. If the method *transform* tries to convert
    a categories which was not seen by method *fit*, it can raise an exception
    or ignore it and replace it by zero.

    .. exref::
        :title: DictVectorizer or CategoriesToIntegers
        :tag: sklearn

        Example which transforms text into integers:

        .. runpython::
            :showcode:

            import pandas
            from mlinsights.mlmodel import CategoriesToIntegers
            df = pandas.DataFrame( [{"cat": "a"}, {"cat": "b"}] )
            trans = CategoriesToIntegers()
            trans.fit(df)
            newdf = trans.transform(df)
            print(newdf)
    """

    def __init__(self, columns=None, remove=None, skip_errors=False, single=False):
        """
        @param      columns         specify a columns selection
        @param      remove          modalities to remove
        @param      skip_errors     skip when a new categories appear (no 1)
        @param      single          use a single column per category, do not multiply them for each value

        The logging function displays a message when a new dense and big matrix
        is created when it should be sparse. A sparse matrix should be allocated instead.
        """
        BaseEstimator.__init__(self)
        TransformerMixin.__init__(self)
        self.columns = columns if isinstance(
            columns, list) or columns is None else [columns]
        self.skip_errors = skip_errors
        self.remove = remove
        self.single = single

    def __str__(self):
        """
        usual
        """
        return self.__repr__()

    def fit(self, X, y=None, **fit_params):
        """
        Makes the list of all categories in input *X*.
        *X* must be a dataframe.

        :param X: iterable
            Training data
        :param y: iterable, default=None
            Training targets.
        :return: self
        """
        if not isinstance(X, pandas.DataFrame):
            raise TypeError(  # pragma: no cover
                "this transformer only accept Dataframes, not {0}".format(type(X)))
        if self.columns:
            columns = self.columns
        else:
            columns = [c for c, d in zip(
                X.columns, X.dtypes) if d in (object,)]

        self._fit_columns = columns
        max_cat = max(len(X) // 2 + 1, 10000)

        self._categories = {}
        for c in columns:
            distinct = set(X[c].dropna())
            nb = len(distinct)
            if nb >= max_cat:
                raise ValueError(  # pragma: no cover
                    "Too many categories ({0}) for one column '{1}' max_cat={2}".format(nb, c, max_cat))
            self._categories[c] = dict((c, i)
                                       for i, c in enumerate(list(sorted(distinct))))
        self._schema = self._build_schema()
        return self

    def _build_schema(self):
        """
        Concatenates all the categories
        given the information stored in *_categories*.

        @return             list of columns, beginning of each
        """
        schema = []
        position = {}
        new_vector = {}
        last = 0
        for c, v in self._categories.items():
            sch = [(_[1], "{0}={1}".format(c, _[1]))
                   for _ in sorted((n, d) for d, n in v.items())]
            if self.remove:
                sch = [d for d in sch if d[1] not in self.remove]
            position[c] = last
            new_vector[c] = {d[0]: i for i, d in enumerate(sch)}
            last += len(sch)
            schema.extend(_[1] for _ in sch)

        return schema, position, new_vector

    def transform(self, X, y=None, **fit_params):
        """
        Transforms categories in numerical features based on the list
        of categories found by method *fit*.
        *X* must be a dataframe. The function does not preserve
        the order of the columns.

        :param X: iterable
            Training data
        :param y: iterable, default=None
            Training targets.
        :return: DataFrame, *X* with categories.
        """
        if not isinstance(X, pandas.DataFrame):
            raise TypeError(  # pragma: no cover
                "X is not a dataframe: {0}".format(type(X)))

        if self.single:
            b = not self.skip_errors

            def transform(v, vec):
                "transform a vector"
                if v in vec:
                    return vec[v]
                elif v is None:
                    return numpy.nan
                elif isinstance(v, float) and numpy.isnan(v):
                    return numpy.nan
                elif not self.skip_errors:
                    lv = list(sorted(vec))
                    if len(lv) > 20:
                        lv = lv[:20]
                        lv.append("...")
                    raise ValueError(  # pragma: no cover
                        "Unable to find category value '{0}' type(v)={2} among\n{1}".format(
                            v, "\n".join(lv), type(v)))
                else:
                    return numpy.nan

            sch, pos, new_vector = self._schema
            X = X.copy()
            for c in self._fit_columns:
                X[c] = X[c].apply(lambda v, cv=c: transform(v, new_vector[cv]))
            return X
        else:
            dfcat = X[self._fit_columns]
            dfnum = X[[c for c in X.columns if c not in self._fit_columns]]
            sch, pos, new_vector = self._schema
            vec = new_vector

            # new_size = X.shape[0] * len(sch)
            res = numpy.zeros((X.shape[0], len(sch)))
            res.fill(numpy.nan)
            b = not self.skip_errors

            for i, row in enumerate(dfcat.to_dict("records")):
                for k, v in row.items():
                    if v is None or (isinstance(v, float) and numpy.isnan(v)):
                        # missing values
                        continue
                    if v not in vec[k]:
                        if b:
                            lv = list(sorted(vec[k]))
                            if len(lv) > 20:
                                lv = lv[:20]
                                lv.append("...")
                            raise ValueError(  # pragma: no cover
                                "unable to find category value '{0}': '{1}' type(v)={3} among\n{2}".format(
                                    k, v, "\n".join(lv), type(v)))
                    else:
                        p = pos[k] + vec[k][v]
                    res[i, p] = 1.0

            if dfnum.shape[1] > 0:
                newdf = pandas.DataFrame(res, columns=sch, index=dfcat.index)
                allnum = pandas.concat([dfnum, newdf], axis=1)
            else:
                allnum = pandas.DataFrame(res, columns=sch, index=dfcat.index)

            return allnum

    def fit_transform(self, X, y=None, **fit_params):
        """
        Fits and transforms categories in numerical features based on the list
        of categories found by method *fit*.
        *X* must be a dataframe. The function does not preserve
        the order of the columns.

        :param X: iterable
            Training data
        :param y: iterable, default=None
            Training targets.
        :return: Dataframe, *X* with categories.
        """
        return self.fit(X, y=y, **fit_params).transform(X, y)



"""
@file
@brief Functions about parameters.
"""
import textwrap


def format_value(v):
    """
    Formats a value to be included in a string.

    @param      v           a string
    @return                 a string
    """
    return ("'{0}'".format(v.replace("'", "\\'"))
            if isinstance(v, str) else "{0}".format(v))


def format_parameters(pdict):
    """
    Formats a list of parameters.

    @param      pdict       dictionary
    @return                 string

    .. runpython::
        :showcode:

        from mlinsights.helpers.parameters import format_parameters

        d = dict(i=2, x=6.7, s="r")
        print(format_parameters(d))
    """
    res = []
    for k, v in sorted(pdict.items()):
        res.append('{0}={1}'.format(k, format_value(v)))
    return ", ".join(res)


def format_function_call(name, pdict):
    """
    Formats a function call with named parameters.

    @param      pdict       dictionary
    @return                 string

    .. runpython::
        :showcode:

        from mlinsights.helpers.parameters import format_function_call

        d = dict(i=2, x=6.7, s="r")
        print(format_function_call("fct", d))
    """
    res = '{0}({1})'.format(name, format_parameters(pdict))
    return "\n".join(textwrap.wrap(res, width=70, subsequent_indent='    '))



"""
@file
@brief Implements a transformer which wraps a predictor
to do transfer learning.
"""
import inspect
from sklearn.base import BaseEstimator, TransformerMixin
from .sklearn_testing import clone_with_fitted_parameters


class TransferTransformer(BaseEstimator, TransformerMixin):
    """
    Wraps a predictor or a transformer in a transformer.
    This model is frozen: it cannot be trained and only
    computes the predictions.

    .. index:: transfer learning, frozen model
    """

    def __init__(self, estimator, method=None, copy_estimator=True,
                 trainable=False):
        """
        @param      estimator           estimator to wrap in a transformer, it is cloned
                                        with the training data (deep copy) when fitted
        @param      method              if None, guess what method should be called,
                                        *transform* for a transformer,
                                        *predict_proba* for a classifier,
                                        *decision_function* if found,
                                        *predict* otherwiser
        @param      copy_estimator      copy the model instead of taking a reference
        @param      trainable           the transfered model must be trained
        """
        TransformerMixin.__init__(self)
        BaseEstimator.__init__(self)
        self.estimator = estimator
        self.copy_estimator = copy_estimator
        self.trainable = trainable
        if method is None:
            if hasattr(estimator, "transform"):
                method = "transform"
            elif hasattr(estimator, "predict_proba"):
                method = "predict_proba"
            elif hasattr(estimator, "decision_function"):
                method = "decision_function"
            elif hasattr(estimator, "predict"):
                method = "predict"
            else:
                raise AttributeError(  # pragma: no cover
                    "Cannot find a method transform, predict_proba, decision_function, "
                    "predict in object {}".format(type(estimator)))
        if not hasattr(estimator, method):
            raise AttributeError(  # pragma: no cover
                "Cannot find method '{}' in object {}".format(
                    method, type(estimator)))
        self.method = method

    def fit(self, X=None, y=None, sample_weight=None):
        """
        The function does nothing.

        :param X: unused
        :param y: unused
        :param sample_weight: unused
        :return: self: returns an instance of self.

        Fitted attributes:

        * `estimator_`: already trained estimator
        """
        if self.copy_estimator:
            self.estimator_ = clone_with_fitted_parameters(self.estimator)
            from .sklearn_testing import assert_estimator_equal  # pylint: disable=C0415
            assert_estimator_equal(self.estimator_, self.estimator)
        else:
            self.estimator_ = self.estimator
        if self.trainable:
            insp = inspect.signature(self.estimator_.fit)
            pars = insp.parameters
            if 'y' in pars and 'sample_weight' in pars:
                self.estimator_.fit(X, y, sample_weight)
            elif 'y' in pars:
                self.estimator_.fit(X, y)
            elif 'sample_weight' in pars:
                self.estimator_.fit(X, sample_weight=sample_weight)
            else:
                self.estimator_.fit(X)
        return self

    def transform(self, X):
        """
        Runs the predictions.

        :param X: numpy array or sparse matrix of shape [n_samples,n_features]
            Training data
        :return: tranformed *X*
        """
        meth = getattr(self.estimator_, self.method)
        return meth(X)



"""
@file
@brief Helpers to visualize a pipeline.
"""
import pprint
from collections import OrderedDict
import numpy
import pandas
from sklearn.base import TransformerMixin, ClassifierMixin, RegressorMixin
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.compose import ColumnTransformer, TransformedTargetRegressor
from ..helpers.pipeline import enumerate_pipeline_models


def _pipeline_info(pipe, data, context, former_data=None):
    """
    Internal function to convert a pipeline into
    some graph.
    """
    def _get_name(context, prefix='-v-', info=None, data=None):
        if info is None:
            raise RuntimeError("info should not be None")  # pragma: no cover
        if isinstance(prefix, list):
            return [_get_name(context, el, info, data) for el in prefix]
        if isinstance(prefix, int):
            prefix = former_data[prefix]
        if isinstance(prefix, int):
            raise TypeError(  # pragma: no cover
                "prefix must be a string.\ninfo={}".format(info))
        sug = "%s%d" % (prefix, context['n'])
        while sug in context['names']:
            context['n'] += 1
            sug = "%s%d" % (prefix, context['n'])
        context['names'][sug] = info
        return sug

    def _get_name_simple(name, data):
        if isinstance(name, str):
            return name
        res = data[name]
        if isinstance(res, int):
            raise RuntimeError(  # pragma: no cover
                "Column name is still a number and not a name: {} and {}."
                "".format(name, data))
        return res

    if isinstance(pipe, Pipeline):
        infos = []
        for _, model in pipe.steps:
            info = _pipeline_info(model, data, context)
            data = info[-1]["outputs"]
            infos.extend(info)
        return infos

    if isinstance(pipe, ColumnTransformer):
        infos = []
        outputs = []
        for _, model, vs in pipe.transformers:
            if all(map(lambda o: isinstance(o, int), vs)):
                new_data = []
                if isinstance(data, OrderedDict):
                    new_data = [_[1] for _ in data.items()]
                else:
                    mx = max(vs)
                    while len(new_data) < mx:
                        if len(data) > len(new_data):
                            new_data.append(data[len(new_data)])
                        else:
                            new_data.append(data[-1])
            else:
                new_data = OrderedDict()
                for v in vs:
                    new_data[v] = data.get(v, v)

            info = _pipeline_info(
                model, new_data, context, former_data=new_data)
            #new_outputs = []
            # for o in info[-1]['outputs']:
            #    add = _get_name(context, prefix=o, info=info)
            #    outputs.append(add)
            #    new_outputs.append(add)
            #info[-1]['outputs'] = new_outputs
            outputs.extend(info[-1]['outputs'])
            infos.extend(info)

        final_hat = False
        if pipe.remainder == "passthrough":

            done = [set(d['inputs']) for d in info]
            merged = done[0]
            for d in done[1:]:
                merged.union(d)
            new_data = OrderedDict(
                [(k, v) for k, v in data.items() if k not in merged])

            info = _pipeline_info(
                "passthrough", new_data, context, former_data=new_data)
            outputs.extend(info[-1]['outputs'])
            infos.extend(info)
            final_hat = True

        if len(pipe.transformers) > 1 or final_hat:
            info = {'name': 'union', 'inputs': outputs, 'type': 'transform'}
            info['outputs'] = [_get_name(context, info=info)]
            infos.append(info)
        return infos

    if isinstance(pipe, FeatureUnion):
        infos = []
        outputs = []
        for _, model in pipe.transformer_list:
            info = _pipeline_info(model, data, context)
            new_outputs = []
            for o in info[-1]['outputs']:
                add = _get_name(context, prefix=o, info=info)
                outputs.append(add)
                new_outputs.append(add)
            info[-1]['outputs'] = new_outputs
            infos.extend(info)
        if len(pipe.transformer_list) > 1:
            info = {'name': 'union', 'inputs': outputs, 'type': 'transform'}
            info['outputs'] = [_get_name(context, info=info)]
            infos.append(info)
        return infos

    if isinstance(pipe, TransformedTargetRegressor):
        raise NotImplementedError(  # pragma: no cover
            "Not yet implemented for TransformedTargetRegressor.")

    if isinstance(pipe, TransformerMixin):
        info = {'name': pipe.__class__.__name__, 'type': 'transform'}
        if len(data) == 1:
            info['outputs'] = data
            info['inputs'] = data
            info = [info]
        else:
            info['inputs'] = [_get_name(context, info=info)]
            info['outputs'] = [_get_name(context, info=info)]
            info = [{'name': 'union', 'outputs': info['inputs'],
                     'inputs': data, 'type': 'transform'}, info]
        return info

    if isinstance(pipe, ClassifierMixin):
        info = {'name': pipe.__class__.__name__, 'type': 'classifier'}
        exp = ['PredictedLabel', 'Probabilities']
        if len(data) == 1:
            info['outputs'] = exp
            info['inputs'] = data
            info = [info]
        else:
            info['outputs'] = exp
            info['inputs'] = [_get_name(context, info=info)]
            info = [{'name': 'union', 'outputs': info['inputs'], 'inputs': data,
                     'type': 'transform'}, info]
        return info

    if isinstance(pipe, RegressorMixin):
        info = {'name': pipe.__class__.__name__, 'type': 'regressor'}
        exp = ['Prediction']
        if len(data) == 1:
            info['outputs'] = exp
            info['inputs'] = data
            info = [info]
        else:
            info['outputs'] = exp
            info['inputs'] = [_get_name(context, info=info)]
            info = [{'name': 'union', 'outputs': info['inputs'], 'inputs': data,
                     'type': 'transform'}, info]
        return info

    if isinstance(pipe, str):
        if pipe == "passthrough":
            info = {'name': 'Identity', 'type': 'transform'}
            info['inputs'] = [_get_name_simple(n, former_data) for n in data]
            if isinstance(data, (OrderedDict, dict)) and len(data) > 1:
                info['outputs'] = [
                    _get_name(context, data=k, info=info)
                    for k in data]
            else:
                info['outputs'] = _get_name(context, data=data, info=info)
            info = [info]
        else:
            raise NotImplementedError(  # pragma: no cover
                "Not yet implemented for keyword '{}'.".format(type(pipe)))
        return info

    raise NotImplementedError(  # pragma: no cover
        "Not yet implemented for {}.".format(type(pipe)))


def pipeline2dot(pipe, data, **params):
    """
    Exports a *scikit-learn* pipeline to
    :epkg:`DOT` language. See :ref:`visualizepipelinerst`
    for an example.

    @param      pipe        *scikit-learn* pipeline
    @param      data        training data as a dataframe or a numpy array,
                            or just a list with the variable names
    @param      params      additional params to draw the graph
    @return                 string

    Default options for the graph are:

    ::

        options = {
            'orientation': 'portrait',
            'ranksep': '0.25',
            'nodesep': '0.05',
            'width': '0.5',
            'height': '0.1',
        }
    """
    raw_data = data
    data = OrderedDict()
    if isinstance(raw_data, pandas.DataFrame):
        for k, c in enumerate(raw_data.columns):
            data[c] = 'sch0:f%d' % k
    elif isinstance(raw_data, numpy.ndarray):
        if len(raw_data.shape) != 2:
            raise NotImplementedError(  # pragma: no cover
                "Unexpected training data dimension: {}.".format(data.shape))
        for i in range(raw_data.shape[1]):
            data['X%d' % i] = 'sch0:f%d' % i
    elif not isinstance(raw_data, list):
        raise TypeError(  # pragma: no cover
            "Unexpected data type: {}.".format(type(raw_data)))

    options = {
        'orientation': 'portrait',
        'ranksep': '0.25',
        'nodesep': '0.05',
        'width': '0.5',
        'height': '0.1',
    }
    options.update(params)

    exp = ["digraph{"]
    for opt in {'orientation', 'pad', 'nodesep', 'ranksep'}:
        if opt in options:
            exp.append("  {}={};".format(opt, options[opt]))
    fontsize = 8
    info = [dict(schema_after=data)]
    names = OrderedDict()
    for d in data:
        names[d] = info
    info.extend(_pipeline_info(pipe, data, context=dict(n=0, names=names)))
    columns = OrderedDict()

    for i, line in enumerate(info):
        if i == 0:
            schema = line['schema_after']
            labs = []
            for c, col in enumerate(schema):
                columns[col] = 'sch0:f{0}'.format(c)
                labs.append("<f{0}> {1}".format(c, col))
            node = '  sch0[label="{0}",shape=record,fontsize={1}];'.format(
                "|".join(labs), params.get('fontsize', fontsize))
            exp.append(node)
        else:
            exp.append('')
            if line['type'] == 'transform':
                node = '  node{0}[label="{1}",shape=box,style="filled' \
                    ',rounded",color=cyan,fontsize={2}];'.format(
                        i, line['name'],
                        int(params.get('fontsize', fontsize) * 1.5))
            else:
                node = '  node{0}[label="{1}",shape=box,style="filled,' \
                    'rounded",color=yellow,fontsize={2}];'.format(
                        i, line['name'],
                        int(params.get('fontsize', fontsize) * 1.5))
            exp.append(node)

            for inp in line['inputs']:
                if isinstance(inp, int):
                    raise IndexError(  # pragma: no cover
                        "Unable to guess columns {} in\n{}\n---\n{}".format(
                            inp, pprint.pformat(columns), '\n'.join(exp)))
                else:
                    nc = columns.get(inp, inp)
                edge = '  {0} -> node{1};'.format(nc, i)
                exp.append(edge)

            labs = []
            for c, out in enumerate(line['outputs']):
                columns[out] = 'sch{0}:f{1}'.format(i, c)
                labs.append("<f{0}> {1}".format(c, out))
            node = '  sch{0}[label="{1}",shape=record,fontsize={2}];'.format(
                i, "|".join(labs), params.get('fontsize', fontsize))
            exp.append(node)

            for out in line['outputs']:
                nc = columns[out]
                edge = '  node{1} -> {0};'.format(nc, i)
                if edge not in exp:
                    exp.append(edge)

    exp.append('}')
    return "\n".join(exp)


def pipeline2str(pipe, indent=3):
    """
    Exports a *scikit-learn* pipeline to text.

    @param      pipe        *scikit-learn* pipeline
    @return                 str

    .. runpython::
        :showcode:

        from sklearn.linear_model import LogisticRegression
        from sklearn.impute import SimpleImputer
        from sklearn.preprocessing import OneHotEncoder
        from sklearn.preprocessing import StandardScaler, MinMaxScaler
        from sklearn.compose import ColumnTransformer
        from sklearn.pipeline import Pipeline

        from mlinsights.plotting import pipeline2str

        numeric_features = ['age', 'fare']
        numeric_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='median')),
            ('scaler', StandardScaler())])

        categorical_features = ['embarked', 'sex', 'pclass']
        categorical_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='constant', fill_value='missing')),
            ('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features),
            ])

        clf = Pipeline(steps=[('preprocessor', preprocessor),
                              ('classifier', LogisticRegression(solver='lbfgs'))])
        text = pipeline2str(clf)
        print(text)
    """
    rows = []
    for coor, model, vs in enumerate_pipeline_models(pipe):
        spaces = " " * indent * (len(coor) - 1)
        if vs is None:
            msg = "{}{}".format(spaces, model.__class__.__name__)
        else:
            v = ','.join(map(str, vs))
            msg = "{}{}({})".format(spaces, model.__class__.__name__, v)
        rows.append(msg)
    return "\n".join(rows)



# -*- coding: utf-8 -*-
"""
@file
@brief Defines class @see cl SkLearnParameters.
"""
import textwrap


class SkException (Exception):

    """
    custom exception
    """
    pass


class SkLearnParameters:

    """
    Defines a class to store parameters of a *learner* or a *transform*.
    """

    def __init__(self, **kwargs):
        """
        Stores parameters as members of the class itself.
        """
        self._keys = list(kwargs.keys())
        for k, v in kwargs.items():
            self.validate(k, v)
            setattr(self, k, v)

    def validate(self, name, value):
        """
        Verifies a parameter and its value.

        @param      name        name
        @param      value       value
        @raises                 raises @see cl SkException if error
        """
        if name.startswith("_") or name.endswith("_"):
            raise SkException(  # pragma: no cover
                "Parameter name must not start by '_': '{0}'".format(name))

    @property
    def Keys(self):
        """
        Returns parameter names.
        """
        return self._keys

    def __repr__(self):
        """
        usual
        """
        def fmt(v):
            "formatting function"
            if isinstance(v, str):
                return "'{0}'".format(v)
            return repr(v)
        text = ", ".join("{0}={1}".format(k, fmt(getattr(self, k)))
                         for k in sorted(self.Keys))
        return "\n".join(textwrap.wrap(text, subsequent_indent="    "))

    def to_dict(self):
        """
        Returns parameters as a dictionary.

        @return         dict
        """
        return {k: getattr(self, k) for k in self.Keys}



# -*- coding: utf-8 -*-
"""
@file
@brief Implements a *transform* which converts a *learner* into
a *transform*.
"""
import textwrap
import numpy
from .sklearn_base_transform import SkBaseTransform


class SkBaseTransformLearner(SkBaseTransform):
    """
    A *transform* which hides a *learner*, it converts
    method *predict* into *transform*. This way,
    two learners can be inserted into the same pipeline.
    There is another a,d shorter implementation
    with class @see class TransferTransformer.

    .. exref::
        :title: Use two learners into a same pipeline
        :tag: sklearn
        :lid: ex-pipe2learner

        It is impossible to use two *learners* into a pipeline
        unless we use a class such as @see cl SkBaseTransformLearner
        which disguise a *learner* into a *transform*.

        .. runpython::
            :showcode:
            :warningout: FutureWarning

            from sklearn.model_selection import train_test_split
            from sklearn.datasets import load_iris
            from sklearn.linear_model import LogisticRegression
            from sklearn.tree import DecisionTreeClassifier
            from sklearn.metrics import accuracy_score
            from sklearn.pipeline import make_pipeline
            from mlinsights.sklapi import SkBaseTransformLearner

            data = load_iris()
            X, y = data.data, data.target
            X_train, X_test, y_train, y_test = train_test_split(X, y)

            try:
                pipe = make_pipeline(LogisticRegression(),
                                     DecisionTreeClassifier())
            except Exception as e:
                print("ERROR:")
                print(e)
                print('.')

            pipe = make_pipeline(SkBaseTransformLearner(LogisticRegression()),
                                 DecisionTreeClassifier())
            pipe.fit(X_train, y_train)
            pred = pipe.predict(X_test)
            score = accuracy_score(y_test, pred)
            print("pipeline avec deux learners :", score)
    """

    def __init__(self, model=None, method=None, **kwargs):
        """
        @param  model   learner instance
        @param  method  method to call to transform the feature (see below)
        @param  kwargs  parameters

        Options for parameter *method*:

        * ``'predict'``
        * ``'predict_proba'``
        * ``'decision_function'``
        * a function

        If *method is None*, the function tries first
        ``predict_proba`` then ``predict`` until one of them
        is part of the class.
        """
        super().__init__(**kwargs)
        self.model = model
        if model is None:
            raise ValueError("value cannot be None")  # pragma: no cover
        if method is None:
            for name in {'predict_proba', 'predict', 'transform'}:
                if hasattr(model.__class__, name):
                    method = name
            if method is None:
                raise ValueError(  # pragma: no cover
                    "Unable to guess a default method for '{0}'".format(repr(model)))
        self.method = method
        self._set_method(method)

    def _set_method(self, method):
        """
        Defines the method to use to convert the features
        into predictions.
        """
        if isinstance(method, str):
            if method == 'predict':
                self.method_ = self.model.predict
            elif method == 'predict_proba':
                self.method_ = self.model.predict_proba
            elif method == 'decision_function':
                self.method_ = self.model.decision_function
            elif method == 'transform':
                self.method_ = self.model.transform
            else:
                raise ValueError(  # pragma: no cover
                    "Unexpected method '{0}'".format(method))
        elif callable(method):
            self.method_ = method
        else:
            raise TypeError(  # pragma: no cover
                "Unable to find the transform method, method={0}".format(method))

    def fit(self, X, y=None, **kwargs):
        """
        Trains a model.

        @param      X               features
        @param      y               targets
        @param      kwargs          additional parameters
        @return                     self
        """
        self.model.fit(X, y=y, **kwargs)
        return self

    def transform(self, X):
        """
        Predictions, output of the embedded learner.

        @param      X   features
        @return         prédictions
        """
        res = self.method_(X)
        if len(res.shape) == 1:
            res = res[:, numpy.newaxis]
        return res

    ##############
    # cloning API
    ##############

    def get_params(self, deep=True):
        """
        Returns the parameters mandatory to clone the class.

        @param      deep        unused here
        @return                 dict
        """
        res = self.P.to_dict()
        res['model'] = self.model
        res['method'] = self.method
        if deep:
            par = self.model.get_params(deep)
            for k, v in par.items():
                res["model__" + k] = v
        return res

    def set_params(self, **values):
        """
        Sets parameters.

        @param      values      parameters
        """
        if 'model' in values:
            self.model = values['model']
            del values['model']
        elif not hasattr(self, 'model') or self.model is None:
            raise KeyError(  # pragma: no cover
                "Missing key '{0}' in [{1}]".format(
                    'model', ', '.join(sorted(values))))
        if 'method' in values:
            self._set_method(values['method'])
            del values['method']
        for k in values:
            if not k.startswith('model__'):
                raise ValueError(  # pragma: no cover
                    "Parameter '{0}' must start with 'model__'.".format(k))
        d = len('model__')
        pars = {k[d:]: v for k, v in values.items()}
        self.model.set_params(**pars)
        if 'method' in values:
            self.method = values['method']
            self._set_method(values['method'])

    #################
    # common methods
    #################

    def __repr__(self):
        """
        usual
        """
        rp = repr(self.model)
        rps = repr(self.P)
        res = "{0}(model={1}, method={2}, {3})".format(
            self.__class__.__name__, rp, self.method, rps)
        return "\n".join(textwrap.wrap(res, subsequent_indent="    "))
